"""
The MidiTrack class enables the translation notes in
a midi file into Apple's SpeechSynthesis DSL.
<center><img src="/assets/img/sun-wavy.png"></img></center>
"""
from typing import List, Optional, Union

from ..constants import SAY_SEGMENT_SILENCE_DURATION
from ..lib import midi
from ..utils import bpm_to_time, frange
from .base import SayObject
from .lyrics import Lyrics
from .note import Note
from .segment import Segment


class MidiTrack(SayObject):
    def __init__(
        self,
        midi_file: str,
        loops: int = 1,
        # start position
        start: Optional[int] = None,
        start_bpm: Optional[Union[float, int]] = 120,
        start_count: Union[str, float, int] = 0,
        start_time_sig: str = "4/4",
        text: Optional[str] = None,
        **kwargs,
    ):
        """
        Generate a melody from a monophonic midifile
        Args:
            midi_file: A path to the midi file
            loops: the number of times to loop the pattern in the midi file.
            # start position
            start: The number of milliseconds of silence to add to the beginning of the track.
            start_bpm: A BPM to use when calculating the number of milliseconds of silence to add to the beginning of the track.
            start_count: A count to use when calculating the number of milliseconds of silence to add to the beginning of the track.
            start_time_sig: A time signature to use when calculating the number of milliseconds of silence to add to the beginning of the track.
            text: The text to to "sing"
        **note_options: Additional options to pass to `Note`.
                        These will affect how each note of the
                        arpeggiator sounds.
        """
        self.midi_file = midi_file
        self.loops = loops

        self.start = start
        if not self.start:
            self.start = bpm_to_time(start_bpm, start_count, start_time_sig)
        self.start_segment_count = (
            int(self.start / SAY_SEGMENT_SILENCE_DURATION) + 1
        )
        self._note_options = kwargs
        self._start_segments = []
        # text / lyrics
        self.lyrics = None
        if text:
            # HACK: add padding to prevent skipping phonemes
            self.lyrics = Lyrics(" . " + text + " . ")

    @property
    def notes(self) -> List[Note]:
        """
        The generated list of `Note` in the MidiTrack.
        """
        _notes = []
        start_at_phoneme = 0
        last_note_length = 0
        for _ in range(0, self.loops):
            for note in midi.process(self.midi_file):
                note_kwargs = {**self._note_options, **note}
                # handle text / phoneme:
                if self.lyrics:
                    phonemes = self.lyrics.get_phonemes(
                        start_at=start_at_phoneme
                    )
                    if len(phonemes) == 0:
                        start_at_phoneme = 0
                        phonemes = self.lyrics.get_phonemes(
                            start_at=start_at_phoneme
                        )
                    note_kwargs["phoneme"] = phonemes
                note = Note(**note_kwargs)
                last_note_length = note.n_segments
                start_at_phoneme += last_note_length
                _notes.append(note)
        return _notes

    @property
    def segments(self) -> List[Segment]:
        """
        The generated list of `Segment` in the Arp.
        """
        return [segment for note in self.notes for segment in note.segments]

    @property
    def n_notes(self) -> int:
        """
        The number of Notes in the Arp.
        """
        return len(self.notes)

    @property
    def n_segments(self) -> int:
        """
        The number of Segments in the Arp.
        """
        return sum([note.n_segments for note in self.notes])

    @property
    def _start_text(self) -> str:
        """
        Add starting silence to the MidiTrack.
        """
        if not self.start:
            return ""
        _start_segments = []
        time_breaks = list(
            frange(0.0, self.start, SAY_SEGMENT_SILENCE_DURATION, 10)
        )[1:]
        for _, total_time in enumerate(time_breaks):
            _start_segments.append(
                Segment(
                    type="silence",
                    velocity=0,
                    duration=SAY_SEGMENT_SILENCE_DURATION,
                )
            )

        if (
            total_time < self.start
            and len(_start_segments) < self.start_segment_count
        ):
            # add final silent step
            dur = self.start - total_time
            _start_segments.append(
                Segment(type="silence", velocity=0, duration=dur)
            )
        return "\n".join(_start_segments)

    def to_text(self):
        """
        Render this MidiTrack as Apple SpeechSynthesis DSL text
        """
        note_texts = (n.to_text() for n in self.notes)
        return self._start_text + "\n" + "\n".join(note_texts)

    def __repr__(self):
        return f"<MidiTrack {self.midi_file}>"



"""
Given a scale and other parameters, generate a sound-font of each note as an .aiff or .wav file.
"""
import click
from midi_utils import ROOT_TO_MIDI, SCALES

from saysynth.cli.options import (adsr_opts, duration_opts, phoneme_opts,
                                  say_opts, segment_opts, text_opt,
                                  velocity_opts, volume_level_opts)
from saysynth.core import Font


def run(**kwargs):
    font = Font(**kwargs)
    font.play(**kwargs)


@click.command()
@click.option(
    "-cs",
    "--scale-start-at",
    type=str,
    default="C3",
    help="Note name/number to start at",
)
@click.option(
    "-ce",
    "--scale-end-at",
    type=str,
    default="G5",
    show_default=True,
    help="Note name/number to end at",
)
@click.option(
    "-c",
    "--scale",
    type=click.Choice([s.lower() for s in SCALES]),
    default="minor",
    show_default=True,
    help="Scale name to use",
)
@click.option(
    "-k",
    "--key",
    type=click.Choice(ROOT_TO_MIDI.keys()),
    default="C",
    show_default=True,
    help="Root note of scale",
)
@click.option(
    "-od",
    "--output-dir",
    default="./",
    type=str,
    show_default=True,
    help="Directory to write to",
)
@click.option(
    "-f",
    "--format",
    type=click.Choice(["wav", "aiff"]),
    default="aiff",
    show_default=True,
    help="Format of each note's file.",
)
@duration_opts
@phoneme_opts
@text_opt
@velocity_opts
@volume_level_opts
@adsr_opts
@segment_opts
@say_opts
def cli(**kwargs):
    """
    Given a scale and other parameters, generate a soundfont
    of each note as an .aiff or .wav file.
    """
    return run(**kwargs)



"""
Utilities for registering processes as files under `~/.saysynth/pid` so they can be dynamically stopped.
<center><img src="/assets/img/spiral.png"></img></center>
"""
import os
import signal
from pathlib import Path
from typing import Dict, List, Optional, Union

import click

from saysynth.cli.colors import blue, green, yellow
from saysynth.constants import DEFAULT_SEQUENCE_NAME
from saysynth.utils import random_track_name

SEQUENCE_PID_LOG = os.path.expanduser("~/.saysynth/pid")
"""
Where to log child pids of parent processes.
These will take the form of `~/.saysynth/pid/{seq}.{track}.{audio_device}.{parent_pid}`
"""


def _read_pid_file(path: str) -> List[int]:
    with open(path, "r") as f:
        return list([int(line.strip()) for line in f.readlines()])


def _append_pid_file(path: str, pid: int) -> None:
    with open(path, "a") as f:
        f.write(str(pid) + "\n")


def _write_pid_file(path: str, pids: List[int]) -> None:
    with open(path, "w") as f:
        f.write("\n".join([str(p) for p in pids]))


def ensure_pid_log() -> None:
    if not os.path.exists(SEQUENCE_PID_LOG):
        os.makedirs(SEQUENCE_PID_LOG)


def _list_pid_file_paths(
    seq: Optional[str] = None,
    track: Optional[str] = None,
    ad: Optional[str] = None,
    parent_pid: Optional[int] = None,
) -> List[Path]:
    """
    List pid file paths by seq, track, and/or pid.

    Args:
        seq: An optional sequence name to filter file paths by
        track: An optional track name to filter file paths by
        ad: An optional audio device to filter file paths by
        parent_pid: An optional parent_pid to filter file paths by
    """
    return Path(SEQUENCE_PID_LOG).glob(
        f'{seq or "*"}.{track or "*"}.{ad or "*"}.{parent_pid or "*"}'
    )


def list_pids() -> List[Dict[str, Union[str, int, List[int]]]]:
    """
    List and parse all pid file paths and lookup the child pids.
    """
    pids = []
    for path in _list_pid_file_paths():
        seq, track, ad, parent_pid = str(path).split("/")[-1].split(".")
        child_pids = lookup_child_pids(seq, track, ad, parent_pid)
        pids.append(
            {
                "seq": seq if seq != DEFAULT_SEQUENCE_NAME else "none",
                "track": track,
                "ad": ad if ad != "None" else "default",
                "parent_pid": parent_pid,
                "child_pids": child_pids,
            }
        )
    return sorted(pids, key=lambda x: x["seq"] + x["track"])


def add_parent_pid(seq: str, track: str, ad: str, parent_pid: int) -> None:
    """
    Associate a pid with a track wihin a sequence.

    Args:
        seq: An optional sequence name to associate with the parent pid
        track: An optional track name to associate with the parent pid
        ad: An optional audio device to associate with the parent pid
        parent_pid: The parent pid
    """
    ensure_pid_log()
    path = f"{SEQUENCE_PID_LOG}/{seq}.{track}.{ad}.{parent_pid}"
    Path(path).touch()
    return path


def rm_parent_pid(
    seq: Optional[str] = None,
    track: Optional[str] = None,
    ad: Optional[str] = None,
    parent_pid: Optional[int] = None,
) -> None:
    """
    Remove pid log files for a seq and/or track

    Args:
        seq: An optional sequence name to remove pids by
        track: An optional track name to remove pids by
        ad: An optional audio device to remove pids by
        parent_pid: An optional parent_pid to remove pids by
    """
    for path in _list_pid_file_paths(seq, track, ad, parent_pid):
        path.unlink()


def add_child_pid(
    child_pid: int, parent_pid: int, parent_pid_file: Optional[str]
) -> None:
    """
    Add a child pid to a parent_pid.

    Args:
        child_pid: The child process to register with the parent.
        parent_pid: The parent pid.
    """
    if parent_pid_file:
        paths = [parent_pid_file]
    else:
        paths = _list_pid_file_paths(parent_pid=parent_pid)
    for path in paths:
        _append_pid_file(path, child_pid)


def rm_child_pid(child_pid: int, parent_pid: int) -> None:
    """
    Remove a child pid from a parent_pid.

    Args:
        child_pid: The child process to dergisister from the parent.
        parent_pid: The parent pid.
    """
    paths = _list_pid_file_paths(parent_pid=parent_pid)
    for path in paths:
        pids = set(_read_pid_file(path))
        pids.remove(child_pid)
        _write_pid_file(path, pids)


def lookup_parent_pids(
    seq: Optional[str] = None,
    track: Optional[str] = None,
    ad: Optional[str] = None,
    parent_pid: Optional[int] = None,
) -> List[int]:
    """
    Lookup all of the parent pids for a seq and/or track.

    Args:
        seq: An optional sequence name to filter parent pids by
        track: An optional track name to filter parent pids by
        ad: An optional audio device to filter parent pids by
        parent_pid: An optional parent_pid to filter parent pids by
    """
    return [
        int(str(path).split(".")[-1])
        for path in _list_pid_file_paths(seq, track, ad, parent_pid)
    ]


def lookup_child_pids(
    seq: Optional[str] = None,
    track: Optional[str] = None,
    ad: Optional[str] = None,
    parent_pid: Optional[int] = None,
) -> List[int]:
    """
    Lookup the child pids for a seq and/or track.

    Args:
        seq: An optional sequence name to filter child pids by
        track: An optional track name to filter child pids by
        ad: An optional audio device to filter child pids by
        parent_pid: An optional parent_pid to filter child pids by
    """
    pids = []
    for path in _list_pid_file_paths(seq, track, ad, parent_pid):
        pids.extend(_read_pid_file(path))
    return list(set(pids))


def lookup_pids(
    seq: Optional[str] = None,
    track: Optional[str] = None,
    ad: Optional[str] = None,
    parent_pid: Optional[int] = None,
) -> None:
    """
    Lookup all pids for a sequence / track.

    Args:
        seq: An optional sequence name to filter pids by
        track: An optional track name to filter pids by
        ad: An optional audio device to filter pids by
        parent_pid: An optional parent_pid to filter pids by
    """
    return lookup_parent_pids(seq, track, ad, parent_pid) + lookup_child_pids(
        seq, track, ad, parent_pid
    )


def stop_child_pids(
    seq: Optional[str] = None,
    track: Optional[str] = None,
    ad: Optional[str] = None,
    parent_pid: Optional[int] = None,
) -> None:
    """
    Stop all the child pids of a parent.

    Args:
        seq: An optional sequence name to stop child pids by
        track: An optional track name to stop child pids by
        ad: An optional audio device to stop child pids by
        parent_pid: An optional parent_pid to stop child pids by
    """
    pids = lookup_child_pids(seq, track, ad, parent_pid)
    for pid in pids:
        try:
            os.kill(pid, signal.SIGTERM)
        except ProcessLookupError:
            pass
    rm_parent_pid(seq, track, ad, parent_pid)


def handle_cli_options(command, **kwargs) -> dict:
    """
    Initialize controller and set cli options.

    Args:
        command: The name of the cli command (eg: `chord`)
    """
    text = kwargs.get('text', None)
    # check for text as filepath.
    if text and os.path.exists(os.path.expanduser(text)):
        with open(text, 'r') as f:
            kwargs['text'] = f.read().strip()
    parent_pid = os.getpid()
    track_name = random_track_name(command, **kwargs)
    add_parent_pid(track_name, command, kwargs.get("audio_device"), parent_pid)
    click.echo(
        f"▶️ {green('starting')} {blue(track_name)} with {yellow('pid')}: {blue(str(parent_pid))}",
        err=True,
    )
    kwargs["parent_pid"] = parent_pid
    return kwargs



"""
The `saysynth.cli.options` module includes shared `click.option` instances for use throughout all commands.

<center><img src="/assets/img/sun-wavy.png"></img></center>

**NOTE**: The documentation for this module is not rendered properly by `pdoc`, so its best to  click "View Source" to scan its contents.
"""
import types
from collections import defaultdict
from typing import Any, Dict, List, Union

import click
import yaml
from midi_utils.arp import STYLES
from midi_utils.constants import CHORDS

from saysynth.constants import (DEFAULT_BPM_TIME_BPM, DEFAULT_BPM_TIME_COUNT,
                                DEFAULT_BPM_TIME_SIG, SAY_ALL_PHONEMES,
                                SAY_DATA_TYPES, SAY_EMPHASIS, SAY_ENDIANNESS,
                                SAY_EXTRA_OPTION_DELIMITER,
                                SAY_PHONEME_CLASSES, SAY_SEGMENT_MAX_DURATION,
                                SAY_TUNED_VOICES, SAY_VOLUME_LEVEL_PER_NOTE,
                                SAY_VOLUME_LEVEL_PER_SEGMENT, SAY_VOLUME_RANGE)
from saysynth.utils import random_track_name, update_dict


def group_options(*options):
    def wrapper(function):
        for option in reversed(options):
            function = option(function)
        return function

    return wrapper


def csv_list(csv: str) -> List[str]:
    """
    A parser for a csv option
    """
    return [v.strip() for v in csv.split(",") if v.strip()]


def csv_int_list(csv: str) -> List[int]:
    """
    A parser for an integer-typed csv option
    """
    return [int(v) for v in csv_list(csv)]


def prepare_options_for_say(input_text: str, **kwargs):
    """
    TODO: Get rid fo this / move to `saysynth.lib.say`?
    """
    # handle some param edge cases
    rp = kwargs.get("randomize_phoneme")
    # for convenience, set the voice option to the one specified
    # in randomize phoneme.
    if rp and ":" in rp:
        kwargs["voice"] = rp.split(":")[0].strip().title()
    kwargs["input_text"] = input_text
    return kwargs


def format_opt_value(name: str, value: Any) -> Any:
    """
    Format an option value given its name and value.
    If the name is part of the common `OPTS`
    set, the click-configured type function will
    be applied, otherwise the value will be returned
    as a string.
    """
    global OPTS
    name = expand_opt_name(name)
    if name in OPTS:
        return OPTS[name]["obj"].type(value)
    return str(value).strip()


def _standardize_opt_name(opt_name: str) -> str:
    """
    Strip leading dashes from a cli option.
    """
    opt_name = opt_name.strip()
    while True:
        if opt_name[0] == "-":
            opt_name = opt_name[1:]
        else:
            break
    return opt_name.lower().replace("-", "_")


def shorten_opt_name(opt_name: str) -> str:
    """
    If an option is present in `OPTS`, return its short_name, otherwise standardize it.
    """
    o = _standardize_opt_name(opt_name)
    return OPTS.get(o, {}).get("short_name", o)


def expand_opt_name(opt_name: str) -> str:
    """
    Strip leading dashes from an option name, lower case and strip.
    and then expand any shortened / non-canonical opts with the global `OPTS` set.
    """
    o = _standardize_opt_name(opt_name)
    return OPTS.get(o, {}).get("full_name", o)


def get_unspecified_opts(
    context: click.Context,
) -> Dict[str, Union[Dict[str, Any], Any]]:
    """
    Get unspecified options from the click.Context and parse their
    accompanying values using the global `OPTS` set.
    """
    opts = {}
    try:
        for i in range(0, len(context.args), 2):
            raw_cli_opt_name = context.args[i]
            raw_cli_opt_val = context.args[i + 1]
            if SAY_EXTRA_OPTION_DELIMITER in raw_cli_opt_name:
                # handle options which are designed to be nested (eg: tracks in a sequence for config_overrides)
                parent_opt_name, child_opt_name = raw_cli_opt_val.split(
                    SAY_EXTRA_OPTION_DELIMITER
                )
                parent_opt_name = _standardize_opt_name(parent_opt_name)
                child_opt_name = expand_opt_name(child_opt_name)
                if parent_opt_name not in opts:
                    opts[parent_opt_name] = {}
                opts[parent_opt_name][child_opt_name] = format_opt_value(
                    child_opt_name, raw_cli_opt_val
                )
            else:
                cli_opt_name = expand_opt_name(raw_cli_opt_name)
                opts[cli_opt_name] = format_opt_value(
                    cli_opt_name, raw_cli_opt_val
                )
    except IndexError:
        pass
    return opts


def set_config_overrides_opt(
    context: click.Context, **kwargs
) -> Dict[str, Any]:
    """
    Combine the config overrides option and additional, unspecified cli options
    """
    cli_config_overrides = get_unspecified_opts(context)
    yaml_config_overrides = yaml.safe_load(
        kwargs.get("config_overrides", "{}")
    )
    kwargs["config_overrides"] = update_dict(
        cli_config_overrides, yaml_config_overrides
    )
    return kwargs


def log_configurations(track_type, **options) -> Dict[str, Any]:
    """
    Log configurations as yaml and exit
    """
    track_name = random_track_name(track_type, **options)
    options.pop("yaml", None)  # remove yaml option
    configs = {"tracks": [{track_name: {"type": track_type, "options": options}}]}
    click.echo(yaml.safe_dump(configs, indent=4))
    return configs


# Duration Options


duration_opt = click.option(
    "-d",
    "--duration",
    default=10000,
    type=int,
    help="The duration of the note in milliseconds.",
)

bpm_opt = click.option(
    "-db",
    "--duration-bpm",
    "duration_bpm",
    default=DEFAULT_BPM_TIME_BPM,
    show_default=True,
    type=float,
    help="The bpm to use when calculating note duration.",
)
count_opt = click.option(
    "-dc",
    "--duration-count",
    "duration_count",
    default=DEFAULT_BPM_TIME_COUNT,
    type=str,
    show_default=True,
    help="The note length to use when calculating note duration (eg: 1/8 or 0.123 or 3)",
)
time_sig_opt = click.option(
    "-dts",
    "--duration-time-sig",
    "duration_time_sig",
    default=DEFAULT_BPM_TIME_SIG,
    type=str,
    show_default=True,
    help="The time signature to use when calculating note duration",
)

"""
CLI options for controlling note duration.
"""
duration_opts = group_options(duration_opt, bpm_opt, count_opt, time_sig_opt)


phoneme_opt = click.option(
    "-ph",
    "--phoneme",
    default="m",
    help=(
        f"One or more valid phoneme to use. Choose from {', '.join(SAY_ALL_PHONEMES)}. "
        "Multiple phonemes can be combined together into one option eg: ''"
    ),
    show_default=True,
    type=csv_list,
)
randomize_phoneme_opt = click.option(
    "-rp",
    "--randomize-phoneme",
    show_default=True,
    type=str,
    help=(
        "Randomize the phoneme for every note. "
        "If `all` is passed, all phonemes will be used. "
        "Alternatively pass a comma-separated list of phonemes (eg 'm,l,n') or a voice and style, eg: Fred:drone. "
        f"Valid voices include: {', '.join(SAY_TUNED_VOICES)}. "
        f"Valid styles include: {', '.join(SAY_PHONEME_CLASSES)}."
    ),
)
randomize_octave_opt = click.option(
    "-ro",
    "--randomize-octave",
    type=csv_int_list,
    required=False,
    default="",
    help="A comma-separate list of octaves to randomly vary between. You can weight certain octaves by providing them multiple times (eg: 0,0,0-1,-1,2 would prefer the root octave first, one octave down second, and two octaves up third.)",
)

"""
CLI options for controlling phonemes.
"""
phoneme_opts = group_options(
    phoneme_opt, randomize_phoneme_opt, randomize_octave_opt
)


# Start Options


randomize_start_opt = click.option(
    "-rt",
    "--randomize-start",
    type=int,
    nargs=2,
    help="Randomize the number of milliseconds to silence to add before the say text. The first number passed in is the minimum of the range, the second is the max.",
)
start_opt = click.option(
    "-t",
    "--start",
    default=None,
    show_default=True,
    type=float,
    help="The number of milliseconds of silence to add before the say text.",
)
start_bpm_opt = click.option(
    "-tb",
    "--start-bpm",
    default=DEFAULT_BPM_TIME_BPM,
    type=float,
    help="The bpm to use when calculating start time",
)
start_count_opt = click.option(
    "-tc",
    "--start-count",
    default=0,
    type=str,
    show_default=True,
    help="The note length to use when calculating start time (eg: 1/8 or 0.123 or 3)",
)
start_time_sig_opt = click.option(
    "-tts",
    "--start-time-sig",
    default=DEFAULT_BPM_TIME_SIG,
    type=str,
    show_default=True,
    help="The time signature to use when calculating start time",
)

"""
CLI options for adding silence to the beginning of a musical passage.
"""
start_opts = group_options(
    randomize_start_opt,
    start_opt,
    start_bpm_opt,
    start_count_opt,
    start_time_sig_opt,
)


# Segment Options

randomize_segments_opt = click.option(
    "-rs",
    "--randomize-segments",
    type=csv_list,
    required=False,
    default="",
    help="Randomize every segment's 'phoneme', 'octave', and/or 'velocity'. Use commas to separate multiple randomization strategies",
)


segment_duration_opt = click.option(
    "-sd",
    "--segment-duration",
    default=SAY_SEGMENT_MAX_DURATION,
    show_default=True,
    type=float,
    help="The duration an individual phoneme",
)
segment_bpm_opt = click.option(
    "-sb",
    "--segment-bpm",
    default=120.0,
    show_default=True,
    type=float,
    help="The bpm to use when calculating phoneme duration",
)
segment_count_opt = click.option(
    "-sc",
    "--segment-count",
    default="1/16",
    type=str,
    show_default=True,
    help="The note length to use when calculating phoneme duration (eg: 1/8 or 0.123 or 3)",
)
segment_time_sig_opt = click.option(
    "-sts",
    "--segment-time-sig",
    default=DEFAULT_BPM_TIME_SIG,
    type=str,
    show_default=True,
    help="The time signature to use when calculating phoneme duration",
)

"""
CLI options for controlling segment generation.
"""
segment_opts = group_options(
    randomize_segments_opt,
    segment_duration_opt,
    segment_bpm_opt,
    segment_count_opt,
    segment_time_sig_opt,
)


# Velocity Options

velocity_opt = click.option(
    "-vl",
    "--velocity",
    type=int,
    show_default=True,
    default=110,
    help="The midi velocity value to use for each note.",
)
velocity_emphasis_opt = click.option(
    "-ve",
    "--velocity-emphasis",
    "emphasis",
    type=int,
    nargs=2,
    show_default=True,
    default=SAY_EMPHASIS,
    help="Two midi velocity values (between 0 and 127) at which to add emphasis to a note/segment",
)
volume_range_opt = click.option(
    "-vr",
    "--volume-range",
    type=float,
    nargs=2,
    show_default=True,
    default=SAY_VOLUME_RANGE,
    help="The min and max volumes (range: 0.0-1.0) to use when mapping from midi velocities",
)
randomize_velocity_opt = click.option(
    "-rv",
    "--randomize-velocity",
    type=int,
    nargs=2,
    help="Randomize a note's velocity by supplying a min and max midi velocity (eg: -rv 40 120)",
)

"""
CLI options for controlling velocities
"""
velocity_opts = group_options(
    velocity_opt,
    velocity_emphasis_opt,
    volume_range_opt,
    randomize_velocity_opt,
)


volume_level_per_segment_opt = click.option(
    "-vps",
    "--render-volume-level-per-segment",
    "volume_level_per_segment",
    default=SAY_VOLUME_LEVEL_PER_SEGMENT,
    type=int,
    show_default=True,
    help="The number of segments per note to render volume tags. Rendering too many can cause random drop-outs, while too few decreases the granularity of ADSR settings.",
)

volume_level_per_note_opt = click.option(
    "-vpn",
    "--render-volume-level-per-note",
    "volume_level_per_note",
    default=SAY_VOLUME_LEVEL_PER_NOTE,
    type=int,
    show_default=True,
    help="The number of notes per sequence to render volume tags. Rendering too many can cause random drop-outs, while too few decreases the granularity of ADSR settings.",
)

"""
CLI options for adjusting the granularity of volume envelopes.
"""
volume_level_opts = group_options(
    volume_level_per_note_opt, volume_level_per_segment_opt
)


attack_opt = click.option(
    "-at",
    "--attack",
    default=0.0,
    show_default=True,
    type=float,
    help="The percentage of the duration it takes to reach the max volume of the note",
)
decay_opt = click.option(
    "-de",
    "--decay",
    default=0.0,
    show_default=True,
    type=float,
    help="The percentage of the duration it takes to reach the sustain volume of the note",
)
sustain_opt = click.option(
    "-su",
    "--sustain",
    default=1.0,
    type=float,
    show_default=True,
    help="The the sustain volume of the note",
)
release_opt = click.option(
    "-re",
    "--release",
    default=0.0,
    type=float,
    show_default=True,
    help="The percentage of the duration it takes to reach the min volume of the note",
)

"""
CLI options for ADSR functionality.
"""
adsr_opts = group_options(
    attack_opt,
    decay_opt,
    sustain_opt,
    release_opt,
)


# Say Options
exec_opt = click.option(
    "-p",
    "--pipe",
    is_flag=True,
    default=False,
    help=(
        "Don't execute the say command and print the text to the console instead. "
        "NOTE: This doesn't work with the `chord` command since this launches "
        "multiple subprocesses and the outputted text will be jumbled. "
        "In order to output the text representation of a chord, use "
        "the `--output-file` option."
    ),
)

rate_opt = click.option(
    "-r",
    "--rate",
    type=int,
    default=70,
    show_default=True,
    help="Rate to speak at (see `man say`)",
)
voice_opt = click.option(
    "-v",
    "--voice",
    type=click.Choice(SAY_TUNED_VOICES),
    default="Fred",
    show_default=True,
    help="Voice to use",
)
input_file_opt = click.option(
    "-i",
    "--input-file",
    type=str,
    help="Filepath to read text input for say from",
    default=None,
)
audio_output_file_opt = click.option(
    "-ao",
    "--audio-output-file",
    type=str,
    help="File to write audio output to",
)
audio_device_opt = click.option(
    "-ad",
    "--audio-device",
    type=str,
    help="Name of the audio device to send the signal to",
)
networks_send_opt = click.option(
    "-ns",
    "--network-send",
    type=str,
    help="Network address to send the signal to",
)
stereo_opt = click.option(
    "-st",
    "--stereo",
    is_flag=True,
    default=False,
    help="Whether or not to generate a stereo signal",
)
endianness_opt = click.option(
    "-en",
    "--endianness",
    type=click.Choice(SAY_ENDIANNESS),
    default="BE",
    help="Whether or not to generate a stereo signal. See say's documentation on data/file formats for more details.",
)
data_type_opt = click.option(
    "-dt",
    "--data-type",
    type=click.Choice(SAY_DATA_TYPES),
    default="I",
    help="One of F (float), I (integer), or, rarely, UI (unsigned integer). See say's documentation on data/file formats for more details.",
)
sample_size_opt = click.option(
    "-ss",
    "--sample-size",
    type=int,
    default=16,
    show_default=True,
    help="Sample size of the signal. When --data-type is 'I', One of 8, 16, 24, 32, 64. When --data-type is 'F', either 32 or 64. See say's documentation on data/file formats for more details.",
)
sample_rate_opt = click.option(
    "-sr",
    "--sample-rate",
    type=int,
    default=22050,
    show_default=True,
    help="Sample rate of the signal (0:22050). See say's documentation on data/file formats for more details.",
)
quality_opt = click.option(
    "-qu",
    "--quality",
    type=int,
    default=127,
    help="Quality of the signal (1:127). See say's documentation on data/file formats for more details.",
    show_default=True,
)
wait_opt = click.option(
    "-w",
    "--wait",
    is_flag=True,
    default=False,
    help="Whether or not to wait for the process to complete.",
)
yaml_opt = click.option(
    "-y",
    "--yaml",
    is_flag=True,
    default=False,
    help="Optionally print these configurations to the console as yaml. This is useful when constructing a sequence.",
)
# progress_bar_opt = click.option(
#     "-pg",
#     "--progress",
#     is_flag=True,
#     default=False,
#     help="Whether or not to display an interactive progress bar",
# )
# interactive_opt = click.option(
#     "-in",
#     "--interactive",
#     is_flag=True,
#     default=False,
#     help="Whether or not to display highlighted text",
# )
# text_color_opt = click.option(
#     "-cf",
#     "--text-color",
#     type=click.Choice(SAY_COLORS),
#     default="white",
#     help="The text color to use when displaying highlighted text",
# )
# bg_color_opt = click.option(
#     "-cb",
#     "--bg-color",
#     type=click.Choice(SAY_COLORS),
#     default="black",
#     help="The background color to use when displaying highlighted text",
# )

output_file_opt = click.option(
    "-o",
    "--output-file",
    type=str,
    help="A filepath to write the generated text to",
)

"""
CLI options for `say`
"""
say_opts = group_options(
    exec_opt,
    wait_opt,
    rate_opt,
    voice_opt,
    audio_output_file_opt,
    audio_device_opt,
    networks_send_opt,
    stereo_opt,
    endianness_opt,
    data_type_opt,
    sample_size_opt,
    sample_rate_opt,
    quality_opt,
    yaml_opt,
    # progress_bar_opt,
    # interactive_opt,
    # text_color_opt,
    # bg_color_opt,
)


# Chord Options

chord_opt = click.option(
    "-c",
    "--chord",
    required=False,
    default="min69",
    type=click.Choice([c.lower() for c in CHORDS.keys()]),
    help="An optional name of a chord to build using the note as root.",
)
chord_notes_opt = click.option(
    "-cn",
    "--chord-notes",
    required=False,
    default="",
    type=csv_int_list,
    help="An optional list of midi numbers to build a chord from based off of the root. For example, the notes '0,3,7' with the root of 'C1' would create a C-minor chord.",
)
chord_velocities_opt = click.option(
    "-cv",
    "--chord-velocities",
    required=False,
    type=csv_int_list,
    help="A comma-separated list of integers (eg: '50,100,127') specifying the midi velocity each note i the chord. The length of this list much match the number of notes in the chord. --volume-range and --velocity-steps also modify this parameter",
)
chord_inversions_opt = click.option(
    "-ci",
    "--chord-inversions",
    "inversions",
    default="",
    required=False,
    type=csv_int_list,
    help="A comma-separated list of integers (eg: '0,1,-1') specifying the direction and amplitude to invert each note. The length of this list much match the number of notes in the chord (post-stack).",
)
chord_stack_opt = click.option(
    "-cs",
    "--chord-stack",
    "stack",
    default=0,
    required=False,
    type=int,
    help="Stack a chord up (eg: '1' or '2') or down (eg: '-1' or '-2').",
)

"""
CLI options for handling chords (`sy arp` + `sy chord`)
"""
chord_opts = group_options(
    chord_opt,
    chord_notes_opt,
    chord_inversions_opt,
    chord_stack_opt,
)

# Arp Options

#
notes_opt = click.option(
    "-ns",
    "--notes",
    required=False,
    default="",
    type=csv_list,
    help="A comma-separated list of note names / midi note numbers to argpeggiate",
)

octaves_opt = click.option(
    "-oc",
    "--octaves",
    required=False,
    default="0",
    type=csv_int_list,
    help="A comma-separated list of octaves to add to the notes",
)

styles_opt = click.option(
    "-sl",
    "--styles",
    required=False,
    default="down",
    type=csv_list,
    help=f"A comma-separated list of styles/sorting algorithms to apply to the notes. This occurs after octaves are added. \nchoose from:\n {', '.join([str(k) for k in STYLES.keys()])}",
)

velocities_opt = click.option(
    "-vl",
    "--velocities",
    required=False,
    default="100",
    show_default=True,
    type=csv_int_list,
    help="A comma-separated list of velocities to apply to the notes, if this list is shorter than the list of notes, a modulo lookup is performed.",
)

loops_opt = click.option(
    "-l",
    "--loops",
    default=None,
    show_default=True,
    type=int,
    help="The number of times to loop the notes in the pattern. If this is set, it will override the '--duration' option.",
)

## beat duration

beat_duration_opt = click.option(
    "-bd",
    "--beat-duration",
    default=None,
    required=False,
    type=int,
    help="The duration of the beat in milliseconds.",
)
beat_bpm_opt = click.option(
    "-bb",
    "--beat-bpm",
    default=DEFAULT_BPM_TIME_BPM,
    type=float,
    show_default=True,
    help="The bpm to use when calculating beat duration.",
)
beat_count_opt = click.option(
    "-bc",
    "--beat-count",
    default=DEFAULT_BPM_TIME_COUNT,
    type=str,
    show_default=True,
    help="The note count to use when calculating beat duration (eg: 1/8 or 0.123 or 3)",
)
beat_time_sig_opt = click.option(
    "-bts",
    "--beat-time-sig",
    default=DEFAULT_BPM_TIME_SIG,
    type=str,
    show_default=True,
    help="The time signature to use when calculating beat duration",
)


note_duration_opt = click.option(
    "-nd",
    "--note-duration",
    default=None,
    required=False,
    type=int,
    help="The duration of a single note in the arp. Defaults to the beat duration",
)
note_bpm_opt = click.option(
    "-nb",
    "--note-bpm",
    "note_bpm",
    default=DEFAULT_BPM_TIME_BPM,
    type=float,
    show_default=True,
    help="The bpm to use when calculating note duration.",
)
note_count_opt = click.option(
    "-nc",
    "--note-count",
    "note_count",
    default=DEFAULT_BPM_TIME_COUNT,
    type=str,
    show_default=True,
    help="The note length to use when calculating note duration (eg: 1/8 or 0.123 or 3)",
)
note_time_sig_opt = click.option(
    "-nts",
    "--note-time-sig",
    default=DEFAULT_BPM_TIME_SIG,
    type=str,
    show_default=True,
    help="The time signature to use when calculating note duration",
)

"""
CLI options specific to `sy arp`.
"""
arp_opts = group_options(
    notes_opt,
    octaves_opt,
    styles_opt,
    velocities_opt,
    loops_opt,
    beat_duration_opt,
    beat_bpm_opt,
    beat_count_opt,
    beat_time_sig_opt,
    note_duration_opt,
    note_bpm_opt,
    note_count_opt,
    note_time_sig_opt,
)

seq_tracks_opt = click.option(
    "-t",
    "--tracks",
    type=csv_list,
    help="A comma-separated list of track names to `play`, `start`, `stop`, or `render`",
)
seq_audio_devices_opt = click.option(
    "-ad",
    "--audio-devices",
    type=csv_list,
    help="A comma-separated list of audio-devices  to `play`, `start`, `stop`, or `render`",
)
seq_output_dir_opt = click.option(
    "-o",
    "--output-dir",
    type=str,
    default="./",
    help="When using `render`, the directory to write audio files of sequence's individual tracks into.",
)
seq_config_overrides_opt = click.option(
    "-c",
    "--config-overrides",
    type=str,
    default="{}",
    help="""
    Override global and track-level configurations at runtime
    by passing in yaml-formatted configurations,
    eg: `-c '{"foo":"bar"}'`.
    These configurations can be specified at the track-level
    by nesting them under the track name,
    eg: `-c '{"track":{"foo":"bar"}}'`.

    You can also override configurations by providing extra command line arguments
    available to `midi`, `note`, `chord`, rand/or `arp` tracks, eg: `-sd 10` or `--segment-duration 10`.
    These can be similarly nested by using a `__` separator, eg: `--track__segment-duration 10`.
    Parameters specified via the --config-overrides option will
    take precedence over any extra CLI arguments.
    """,
)
seq_command_arg = click.argument(
    "command",
    type=click.Choice(["play", "start", "stop", "render", "echo"]),
    required=True,
)

"""
CLI options specific to `sy seq` and `sy demo`.
"""
seq_opts = group_options(
    seq_tracks_opt,
    seq_audio_devices_opt,
    seq_output_dir_opt,
    seq_config_overrides_opt,
)

"""
Text to to use when selecting phonemes; the text to 'sing'
"""
text_opt = click.option(
    "-tx",
    "--text",
    type=str,
    default=None,
    help="Text to to use when selecting phonemes; the text to 'sing'. Can also be a path to a file containing text to sing.",
)


def _build_option_set(locals) -> Dict[str, click.Parameter]:
    """
    Meta-programming HACK to build an option set we can use to dynamically
    parse extra cli options.
    """
    click_params = defaultdict(dict)
    for _, object in locals.items():
        if not (
            isinstance(object, types.FunctionType)
            and object.__module__.startswith("click")
        ):
            continue
        # instantiate this option and fetch its param object
        param: click.Parameter = object(lambda x: x).__click_params__[0]
        full_name = _standardize_opt_name(param.name)
        option = {"obj": param, "full_name": full_name}
        click_params[param.name] = option

        # also add lookup for short name flags
        short_cli_opt = [
            o
            for o in param.opts
            if o.startswith("-") and not o.startswith("--")
        ]
        if not len(short_cli_opt):
            continue

        # add lookup to short name
        short_name = _standardize_opt_name(short_cli_opt[0])
        click_params[param.name]["short_name"] = short_name
        # also add reverse lookup
        click_params[short_name] = option
        click_params[short_name]["short_name"] = short_name
    return click_params


OPTS: Dict[str, click.Parameter] = _build_option_set(locals())



"""
Synthesize a melody from a monophonic midi file.
"""

import os

import click

from saysynth.cli.options import (adsr_opts, group_options, log_configurations,
                                  phoneme_opts, randomize_velocity_opt,
                                  say_opts, segment_opts, start_opts, text_opt,
                                  velocity_emphasis_opt, volume_level_opts,
                                  volume_range_opt)
from saysynth.constants import DEFAULT_SEQUENCE_NAME
from saysynth.core import MidiTrack, controller


def run(**kwargs):
    midi_track = MidiTrack(**kwargs)
    midi_track.cli(**kwargs)


@click.command()
@click.argument("midi_file", required=True)
@click.option(
    "-l",
    "--loops",
    default=1,
    show_default=True,
    type=int,
    help="The number of times to loop the midi file",
)
@start_opts
@phoneme_opts
@text_opt
@group_options(velocity_emphasis_opt, volume_range_opt, randomize_velocity_opt)
@volume_level_opts
@adsr_opts
@segment_opts
@click.option(
    "-o",
    "--output-file",
    type=str,
    help="A filepath to write the generated text to",
)
@say_opts
def cli(**kwargs):
    """
    Synthesize a melody from a fully-monophonic midi file.
    """
    if kwargs["yaml"]:
        return log_configurations("midi", **kwargs)
    parent_pid = os.getpid()
    ad = kwargs.get("audio_device", "")
    controller.add_parent_pid(DEFAULT_SEQUENCE_NAME, "midi", ad, parent_pid)
    kwargs["parent_pid"] = parent_pid
    return run(**kwargs)



"""
The Word class contains info on the text, number of syllables,
and list of phonemes in a word. It's used in the `Lyric` class.

<center><img src="/assets/img/nuclear.png"></img></center>
"""
import math
from typing import List


class Word(object):
    def __init__(self, text: str, syllables: int, phonemes: List[str]):
        """
        The Word class contains info on the text, number of syllables,
        and list of phonemes in a word. It's used in the `Lyric` class.
        Args:
            text: The raw text of the word
            syllables: The number of syllables in the word
            phonemes: The list of phonemes in the word
        """
        self.text = text
        self.syllables = syllables
        self.phonemes = phonemes
        self.n_phonemes = len(self.phonemes)
        self.max_phonemes_per_syllable = math.floor(
            self.n_phonemes / self.syllables
        )



"""
Assorted utilities for use throughout `saysynth`
"""

import collections
import math
import os
import tempfile
from typing import Any, Dict, List, Union

from hashids import Hashids

from .constants import (DEFAULT_BPM_TIME_BPM, DEFAULT_BPM_TIME_COUNT,
                        DEFAULT_BPM_TIME_SIG)

hashids = Hashids(salt="saysynth", min_length=3, alphabet="42069iacabplurwtf")


def frange(start: float, stop: float, by: float, sig_digits: int = 5):
    """
    Generate a range of float values.

    Args:
        start: The starting value of the range.
        stop: The ending value of the range.
        by: the amount to divide the range by.
        sig_digits:  The number of significant digits to use when rounding.

    """
    div = math.pow(10, sig_digits)
    for value in range(int(start * div), int(stop * div), int(by * div)):
        yield round(value / div, sig_digits)


def here(f, *args):
    """
    Pass `__file__` to get the current directory and `*args` to generate a filepath relative
    to the current directory.

    Args:
        f: Usually `__file__`
    """
    return os.path.join(os.path.dirname(os.path.abspath(f)), *args)


def make_tempfile(format: str = "txt"):
    """
    Make a tempfile
    Args:
        format: The file's suffix.
    """
    return tempfile.mkstemp(suffix=f".{format}")[-1]


def bpm_to_time(
    bpm: float = 120.00,
    count: Union[str, int, float] = 1,
    time_sig: str = DEFAULT_BPM_TIME_SIG,
) -> float:
    """
    Take a bpm, note count, and time_sig and return a length in milliseconds
    Args:
        bpm: The bpm as a float
        count: the count as a string, int, or float (eg: '2/1', 2, 2.0 )
        time_sig: The time signature as a string (eg: '4/4')

    """
    if isinstance(count, str):
        if "/" in count:
            numerator, denominator = count.split("/")
            count = float(numerator) / float(denominator)
    time_segments = time_sig.split("/")
    return (
        (60.00 / float(bpm)) * float(time_segments[0]) * float(count) * 1000.0
    )


def rescale(
    x: Union[int, float],
    range_x: List[Union[int, float]],
    range_y: List[Union[int, float]],
    sig_digits: int = 3,
) -> float:
    """
    Rescale value `x` to scale `y` give the range of `x` and the range of `y`

    Args:
        x: An value to rescale
        range_x: The range ([min, max]) of the origin scale
        range_y: The range ([min, max]) of the target scale
        sig_digits: The number of significant digits to use when rounding.

    """
    # Figure out how 'wide' each range is
    x_min = min(range_x)
    y_min = min(range_y)
    x_span = max(range_x) - x_min
    y_span = max(range_y) - y_min

    # Compute the scale factor between left and right values
    scale_factor = float(y_span) / float(x_span)

    return round(y_min + (x - x_min) * scale_factor, sig_digits)


def calculate_total_duration(**kwargs) -> str:
    """
    Calculate the total duration of a track/chord/etc and
    format it as a string.
    Args:
        start_bpm/count/time_sig: the start of the track
        duration_bpm/count/time_sig: the duration of the passage
    Returns: float
    """
    if not kwargs.get("start_duration", None):
        start_duration = bpm_to_time(
            kwargs.get("start_bpm", DEFAULT_BPM_TIME_BPM),
            kwargs.get("start_count", DEFAULT_BPM_TIME_COUNT),
            kwargs.get("start_time_sig", DEFAULT_BPM_TIME_SIG),
        )
    else:
        start_duration = kwargs.get("start_duration", 0)
    if not kwargs.get("duration", None):
        duration = bpm_to_time(
            kwargs.get("duration_bpm", DEFAULT_BPM_TIME_BPM),
            kwargs.get("duration_count", DEFAULT_BPM_TIME_COUNT),
            kwargs.get("duration_time_sig", DEFAULT_BPM_TIME_SIG),
        )
    else:
        duration = kwargs.get("duration", 0)

    seconds = (start_duration + duration) / 1000.0
    minutes = math.floor(seconds / 60.0)
    remainder_seconds = seconds - (minutes * 60.0)
    return f"{minutes}:{int(math.floor(remainder_seconds)):02}"


def update_dict(d: Dict[Any, Any], u: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Recursively update a dictionary.
    Args:
        d: the dictionary to be updated
        u: the dictionary to use in update
    Returns: dict
    """
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            d[k] = update_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def random_track_name(track_type: str, **options) -> str:
    """
    Generate a unique track name given a type and a dict of options
    """
    hash_id = hashids.encode(
        int.from_bytes((track_type + str(options)).encode(), "little")
    )[:5]
    return f"{track_type}-{str(hash_id)}"



from saysynth.utils import (bpm_to_time, calculate_total_duration,
                            random_track_name)


def test_bpm_to_time_str_count():
    t = bpm_to_time(60, "1/16")
    assert t == 250


def test_bpm_to_time_str_count_time_sig():
    t = bpm_to_time(60, "1/8", "3/3")
    assert t == 375


def test_bpm_to_time_float_count():
    t = bpm_to_time(60, 1.0 / 8.0, "3/3")
    assert t == 375


def test_calculate_total_duration_without_bpm():
    total_duration = calculate_total_duration(
        start_duration=60000, duration=60000
    )
    assert total_duration == "2:00"


def test_calculate_total_duration_with_bpm():
    total_duration = calculate_total_duration(
        start_bpm=120,
        start_count=10,
        start_time_sig="4/4",
        duration_bpm=120,
        duration_count=20,
        duration_time_sig="4/4",
    )
    assert total_duration == "1:00"


def test_calculate_total_duration_with_duration_bpm_and_without_start_bpm():
    total_duration = calculate_total_duration(
        start_duration=30000,
        duration_bpm=120,
        duration_count=15,
        duration_time_sig="4/4",
    )
    assert total_duration == "1:00"


def test_calculate_total_duration_without_duration_bpm_and_with_start_bpm():
    total_duration = calculate_total_duration(
        start_bpm=120, start_count=15, start_time_sig="4/4", duration=30000
    )
    assert total_duration == "1:00"


def test_random_track_name():
    track, hash = random_track_name("test", foo="123").split("-")
    assert track == "test"
    assert len(hash) > 3



"""
Print the current version of `saysynth` to the console.
"""
import click

from saysynth.cli.colors import blue, green, red, yellow
from saysynth.version import VERSION


@click.command()
def cli():
    """
    Print the current version of `saysynth` to the console.
    """
    click.echo(
        f"➡️ {green('saysynth')} ({red('sy')}) {yellow('version:')} {blue(VERSION)}",
        err=True,
    )



"""

`saysynth.core` contains Classes for generating and musical passages in Apple's SpeechSynthesis DSL.

<center><img src="/assets/img/coffee.png"></img></center>

"""
from .arp import Arp  # noqa: F401
from .chord import Chord  # noqa: F401
from .font import Font  # noqa: F401
from .lyrics import Lyrics  # noqa: F401
from .midi_track import MidiTrack  # noqa: F401
from .note import Note  # noqa: F401
from .segment import Segment  # noqa: F401
from .word import Word  # noqa: F401



"""
Utilities for extracing phonemes and syllables from english text.
"""

from functools import lru_cache
from typing import List, Tuple

import nltk
from g2p_en import G2p
from nltk.corpus import cmudict
from nltk.tokenize import word_tokenize

# ensure NLTK resources are downloaded
try:
    cmudict.ensure_loaded()
except LookupError:
    nltk.download("cmudict")
try:
    word_tokenize("test")
except Exception:
    nltk.download("punkt")


from ..constants import G2P_PHONEMES_TO_SAY_PHONEMES

G2P = None
CMU = None
VOWELS = "aeiouy"


@lru_cache(maxsize=1024)
def word_to_g2p_phonemes(text: str) -> List[str]:
    """
    Extract g2p phonemes from a text.

    Args:
        text: A text to extract phonemes from
    """
    global G2P
    if not G2P:
        G2P = G2p()
    return G2P(text)


def word_to_say_phonemes(text: str) -> List[str]:
    """
    Extract Apple SpeechSynthesis phonemes from a text.

    Args:
        text: A text to extract phonemes from
    """
    return [
        G2P_PHONEMES_TO_SAY_PHONEMES.get(p, "")
        for p in word_to_g2p_phonemes(text)
    ]


@lru_cache(maxsize=1024)
def word_to_syllable_count(word: str) -> int:
    """
    Compute the syllable count of a word, using CMU if the word is present in the
    corpus, otherwise [follow this implementation](https://stackoverflow.com/questions/14541303/count-the-number-of-syllables-in-a-word).

    Args:
        word: A word to compute syllables for.
    """
    global CMU
    if not CMU:
        CMU = cmudict.dict()
    try:
        return [
            len(list(y for y in x if y[-1].isdigit()))
            for x in CMU[word.lower()]
        ][0]
    except KeyError:
        # if word not found in cmudict
        # referred from stackoverflow.com/questions/14541303/count-the-number-of-syllables-in-a-word
        count = 0
        word = word.lower()
        if word[0] in VOWELS:
            count += 1
        for index in range(1, len(word)):
            if word[index] in VOWELS and word[index - 1] not in VOWELS:
                count += 1
        if word.endswith("e"):
            count -= 1
        if word.endswith("le"):
            count += 1
        if count == 0:
            count += 1
        return count


@lru_cache(maxsize=1024)
def process_text_for_say(text: str) -> List[Tuple[str, int, List[str]]]:
    """
    Get a list of phonemes + syllable counts for each word in a text.
    Args:
        text: A text to process.
    """
    return [
        (word, word_to_syllable_count(word), word_to_say_phonemes(word))
        if word not in [",", ".", "?", "!", "-", ":", ";"]
        else ("", 1, ["%"])  # silence
        for word in word_tokenize(text)
    ]



"""
The Font class creates a list of `Notes` within a scale and writes them to separate files. This makes them easy to import into samplers or DAWs.
<center><img src="/assets/img/sun-wavy.png"></img></center>
"""
import os
from typing import Union

from midi_utils import midi_scale, midi_to_note, note_to_midi

from .note import Note


class Font(object):
    def __init__(
        self,
        key: Union[str, int],
        scale: str,
        scale_start_at: Union[int, str],
        scale_end_at: Union[int, str],
        **note_options,
    ):
        """
        The Font class creates a list of `Notes` within a
        scale and writes them to separate files. This
        makes them easy to import into samplers or DAWs.
        Scale generation is provided by `midi_scale`
        Args:
            key: The root note of the scale, eg: C,D,G#,etc
            scale: The name of the scale (see midi_utils.constants.SCALES)
            scale_start_at: A note name or midi note number to start the scale at
            scale_end_at: A note name or midi note number to end the scale at
            **note_options: Additional options to pass to `Note` generation.
        """
        # add note type to simplify function call
        self.scale = midi_scale(
            key=key,
            scale=scale,
            min_note=note_to_midi(scale_start_at),
            max_note=note_to_midi(scale_end_at),
        )
        note_options.setdefault("type", "note")
        self.note_options = note_options

    def _get_kwargs(self, midi, kwargs) -> dict:
        kw = dict(self.note_options)
        kw.update(kwargs)
        kw["type"] = "note"
        kw["midi"] = midi
        kw["note"] = midi_to_note(midi)
        return kw

    def play(self, **kwargs) -> None:
        """
        Alias of `self.generate`
        """
        return self.generate(**kwargs)

    def generate(self, **kwargs) -> None:
        # generate files for each note in the scale
        for midi in self.scale:
            kwargs = self._get_kwargs(midi, kwargs)

            # generate audio output file name
            filepath = (
                f"{kwargs['voice'].lower()}-{kwargs['rate']}-"
                f"{kwargs['note']}-{midi:02d}.{kwargs['format']}"
            )
            audio_output_file = os.path.join(
                kwargs["output_dir"],
                filepath,
            )
            # generate input file of text
            note = Note(**kwargs)
            note.play(
                voice=kwargs["voice"],
                rate=kwargs["rate"],
                audio_output_file=audio_output_file,
                wait=True,
            )
            if not os.path.exists(audio_output_file):
                raise RuntimeError(
                    f"File {audio_output_file} was not successfully created"
                )



"""
Helpers for adding colored log output to the CLI.
<center><img src="/assets/img/spiral.png"></img></center>

"""
from click import style


def red(text):
    """
    Red text
    """
    return style(text, fg="red")


def green(text):
    """
    Green text
    """
    return style(text, fg="green")


def yellow(text):
    """
    Yellow text
    """
    return style(text, fg="yellow")


def blue(text):
    """
    Blue text
    """
    return style(text, fg="cyan")



from saysynth.cli.options import (expand_opt_name, format_opt_value,
                                  log_configurations, shorten_opt_name)


def test_expand_opt_name_text():
    assert expand_opt_name("-tx") == "text"
    assert expand_opt_name("tx") == "text"
    assert expand_opt_name("--text") == "text"
    assert expand_opt_name(" --TEXT ") == "text"


def test_shorten_opt_name_text():
    assert shorten_opt_name("-tx") == "tx"
    assert shorten_opt_name("tx") == "tx"
    assert shorten_opt_name("--text") == "tx"
    assert shorten_opt_name(" --TEXT ") == "tx"


def test_format_opt_value_chord_notes():
    assert format_opt_value("-cn", "1,2,4,5") == [1, 2, 4, 5]
    assert format_opt_value("--chord_notes", "1,2,4,5") == [1, 2, 4, 5]


def test_log_configurations():
    configs = log_configurations("chord", yaml=True, voice="Fred")
    config_options = list(configs["tracks"][0].values())[0]["options"]
    assert "yaml" not in config_options
    assert config_options["voice"] == "Fred"



"""
Synthesize an individual note.
"""
import click
from midi_utils import note_to_midi

from saysynth.cli.options import (adsr_opts, duration_opts, log_configurations,
                                  output_file_opt, phoneme_opts, say_opts,
                                  segment_opts, start_opts, text_opt,
                                  velocity_opts, volume_level_per_segment_opt)
from saysynth.core import Note, controller


def run(**kwargs):
    note = Note(**kwargs)
    note.cli(**kwargs)


@click.command()
@click.argument("root", type=note_to_midi, default="A2")
@text_opt
@start_opts
@duration_opts
@phoneme_opts
@velocity_opts
@volume_level_per_segment_opt
@adsr_opts
@segment_opts
@output_file_opt
@say_opts
def cli(**kwargs):
    """
    Generate an individual note.
    """
    if kwargs["yaml"]:
        return log_configurations("note", **kwargs)
    kwargs = controller.handle_cli_options("note", **kwargs)
    return run(**kwargs)



"""The Arp class enables melodic speech synthesis by mapping input text or phonemes onto a configurable arpeggiator.
<center><img src="/assets/img/cell.png"></img></center>
"""
import random
from functools import cached_property
from typing import List, Optional, Union

from midi_utils import midi_arp

from .base import SayObject
from .lyrics import Lyrics
from .note import Note
from .segment import Segment


class Arp(SayObject):
    def __init__(
        self,
        text: Optional[str] = None,
        notes: List[int] = [],
        root: str = "C2",
        chord: str = "min69",
        inversions: List[int] = [],
        stack: int = 0,
        styles: List[str] = ["down"],
        octaves: List[int] = [0],
        velocities: List[int] = [100],
        volume_level_per_note: int = 3,
        beat_bpm: float = 131.0,
        beat_count: Union[float, int, str] = "1/16",
        beat_time_sig: str = "4/4",
        beat_duration: Optional[float] = None,
        note_bpm: float = 131.0,
        note_count: Union[float, int, str] = "3/64",
        note_time_sig: str = "4/4",
        note_duration: Optional[float] = None,
        randomize_start: Optional[List[int]] = None,
        start_bpm: float = 131.0,
        start_count: Union[float, int, str] = 0,
        start_time_sig: str = "4/4",
        start_duration: Optional[float] = None,
        duration_bpm: float = 131.0,
        duration_count: Union[float, int, str] = "16",
        duration_time_sig: str = "4/4",
        duration: Optional[float] = None,
        loops: Optional[int] = None,
        **note_options,
    ):
        """
        Generate an arpeggiated melody

        Args:
            text: text to 'sing'
            notes: an arbitrary list of notes to arpeggiate
            root: root note of chord to arpeggiate
            chord: the chord name
            inversions: a list of inversions to apply to the chord notes
            stack: Stack a chord up (1) or down (-1)
            styles: A list of arpeggiated style names.
                    See https://gitlab.com/gltd/midi-utils/-/blob/main/midi_utils/arp.py for the full list.
            octaves: A list of octaves to add to the notes (eg: [-1, 2])
            velocities: A list of velocities for each note of the arpeggiator.
                        If this list is shorter than the list of notes, a modulo
                        operator is user.
            volume_level_per_note: How many notes it takes to re-adjust the volume/velocity.
                                  This parameter exists because too many volume changes
                                  can cause sporadically audio to dropout.
            beat_bpm: The bpm to use when calculating the duration of each beat of the arp.
            beat_count: The count of one beat of the arp
            beat_time_sig: str = The time signature of the arp.
            beat_duration: The duration of the beat. If provided, this overrides
                           the other beat_* parameters.
            note_bpm: The bpm to use when calculating the duration of each note of the arp.
                      By default, this is the
            note_count: The count of one beat of the arp
            note_time_sig: str = The time signature of the arp.
            note_duration: The duration of the beat. If provided, this overrides
                           the other note_* parameters.
            randomize_start: Optional[List[int]] = None,
            start_bpm: bpm to use when determining the start of the arp. The start_* parameters
                       adds silence at the beginning of the arpeggiator. This is particularly
                       useful when creating sequences.
            start_count: The count of one beat of starting silence.
            start_time_sig: The time signature to use when calculating the duration
                           of one beat of starting silence.
            start_duration: The amount of silence to add at the beginning in ms
            duration_bpm: bpm to use when determining the duration of the arpeggiator.
            duration_count: The duration beat count
            duration_time_sig: Time signature to use when determining duration
            duration: The total duration of the pattern in ms.
            loops: The number of times to loop the pattern.
                   This overrides `duration_*` settings.
            **note_options: Additional options to pass to `Note`.
                          These will affect how each note of the
                          arpeggiator sounds.

        """
        self.styles = styles

        if randomize_start:
            start_duration = random.choice(
                range(randomize_start[0], randomize_start[1] + 1)
            )

        self.sequence = midi_arp(
            notes=notes,
            root=root,
            chord=chord,
            inversions=inversions,
            stack=stack,
            octaves=octaves,
            styles=styles,
            velocities=velocities,
            beat_bpm=beat_bpm,
            beat_count=beat_count,
            beat_time_sig=beat_time_sig,
            beat_duration=beat_duration,
            note_bpm=note_bpm,
            note_count=note_count,
            note_time_sig=note_time_sig,
            note_duration=note_duration,
            start_bpm=start_bpm,
            start_count=start_count,
            start_time_sig=start_time_sig,
            start_duration=start_duration,
            duration_bpm=duration_bpm,
            duration_count=duration_count,
            duration_time_sig=duration_time_sig,
            duration=duration,
            loops=loops,
        )
        self.volume_level_per_note = volume_level_per_note
        self._note_options = note_options
        self.lyrics = None
        if text:
            # HACK: add padding to prevent skipping phonemes
            # TODO: figure why this is happening.
            self.lyrics = Lyrics(" . " + text + " . ")

    def _get_kwargs(self, index, **kwargs):
        """
        get kwargs + update with new ones
        used for mapping similar kwargs over different notes
        """
        d = dict(self._note_options.items())
        d.update(kwargs)
        d["include_volume_level"] = index % self.volume_level_per_note == 0
        return d

    @cached_property
    def notes(self) -> List[Note]:
        """
        The generated list of `Note` in the Arp.
        """
        start_at_phoneme = 0
        _notes = []
        for i, note in enumerate(self.sequence):
            note_kwargs = self._get_kwargs(i, **note)
            if self.lyrics:
                # handle text / phoneme:
                phonemes = self.lyrics.get_phonemes(start_at=start_at_phoneme)
                if len(phonemes) == 0:
                    # TODO: Fix this hack.
                    start_at_phoneme = 0
                    phonemes = self.lyrics.get_phonemes(
                        start_at=start_at_phoneme
                    )
                note_kwargs["phoneme"] = phonemes
            note = Note(**note_kwargs)
            last_note_length = note.n_segments
            start_at_phoneme += last_note_length
            _notes.append(note)
        return _notes

    @property
    def segments(self) -> List[Segment]:
        """
        The generated list of `Segment` in the Arp.
        """
        return [segment for note in self.notes for segment in note.segments]

    @property
    def n_notes(self) -> int:
        """
        The number of Notes in the Arp.
        """
        return len(self.notes)

    @property
    def n_segments(self) -> int:
        """
        The number of Segments in the Arp.
        """
        return sum([note.n_segments for note in self.notes])

    def to_text(self):
        """
        Render this Arp as Apple SpeechSynthesis DSL text.
        """
        return "\n".join([n.to_text() for n in self.notes])

    def __repr__(self):
        return f"<Arp {','.join(self.styles)} {','.join(str(n) for n in self.notes)}>"



"""
Play a sequence of `chord`, `midi`, `note`, and/or `arp` tracks via yaml configuration.
"""

import copy
import os
from argparse import ArgumentError
from multiprocessing.pool import ThreadPool as Pool
from typing import Dict

import click
import yaml

from saysynth.cli.colors import blue, green, red, yellow
from saysynth.cli.commands import arp, chord, midi, note
from saysynth.cli.options import (expand_opt_name, seq_command_arg, seq_opts,
                                  set_config_overrides_opt)
from saysynth.core import controller
from saysynth.utils import calculate_total_duration


def run(**kwargs):
    controller.ensure_pid_log()

    # process configurations
    config_path = kwargs.pop("base_config")
    command = kwargs.pop("command")
    output_dir = kwargs.pop("output_dir", "./")
    if config_path:
        # print config
        if command == "echo":
            click.echo(str(config_path.read()))
            return
        base_config = yaml.safe_load(config_path)
    else:
        base_config = {"name": "*"}

    globals = base_config.pop("globals", {})
    seq = base_config.pop("name", None)
    if seq is None:
        raise ArgumentError("You must set a `name` in your sequence config")

    tracks = kwargs.get("tracks", None) or []
    audio_devices = kwargs.get("audio_devices", None) or []

    # log sequence name
    click.echo("-" * 79, err=True)
    click.echo(f"{red('sequence')}: {yellow(seq)}", err=True)
    click.echo("-" * 79, err=True)

    # play/start sequences/tracks
    if command in ["play", "start", "render"]:
        track_configs = []
        for track, base_track_config in base_config.get("tracks", {}).items():

            # optionally skip tracks
            if len(tracks) and track not in tracks:
                continue

            # allow for track-specific overrides
            config_overrides = kwargs.get("config_overrides", {})
            track_overrides = config_overrides.pop(track, {})
            config_overrides.update(track_overrides)

            # create track config
            track_options = copy.copy(globals)  # start with globals
            base_track_options = {
                expand_opt_name(k): v
                for k, v in base_track_config.get("options", {}).items()
            }  # expand the names of shortened sequence options
            track_options.update(base_track_options)
            track_options.update(config_overrides)

            # optionally filter by audio device
            ad = track_options.get("audio_device", None)
            if len(audio_devices) and ad not in audio_devices:
                continue

            # optionally start tracks immediately
            if command == "start":
                track_options["start_count"] = 0

            # optionally render tracks
            if command == "render":
                try:
                    os.makedirs(output_dir)
                except FileExistsError:
                    pass
                track_options["audio_output_file"] = os.path.join(
                    output_dir,
                    f"{seq}-{track}-{base_track_config['type']}.aiff",
                )

            # construct track configs
            base_track_config["options"] = track_options
            track_configs.append((seq, track, ad, base_track_config, command))

        # run everything
        for _ in Pool(len(track_configs)).imap(run_track_func, track_configs):
            continue

    if command == "stop":

        if len(audio_devices):
            for ad in audio_devices:
                click.echo(
                    f"🛑 {red('stopping')} ➡️ all tracks on  {yellow('audio_device')}: {blue(ad)}",
                    err=True,
                )
                controller.stop_child_pids(seq, track=None, ad=ad)

        else:
            if not len(tracks):
                tracks = ["*"]
            for track in tracks:
                click.echo(
                    f"🛑 {red('stopping')} ➡️ {yellow('track')}: { blue(track)}",
                    err=True,
                )
                controller.stop_child_pids(seq, track)


TRACK_FUNCS: Dict[str, callable] = {
    "chord": chord.run,
    "midi": midi.run,
    "note": note.run,
    "arp": arp.run,
}


def run_track_func(item):
    seq_name, track, ad, kwargs, command = item
    pid = os.getpid()
    parent_pid_file = controller.add_parent_pid(seq_name, track, ad, pid)
    type = kwargs.get("type", None)
    options = kwargs.get("options", {})
    options["parent_pid"] = pid  # pass parent id to child process.
    options["parent_pid_file"] = parent_pid_file

    if type not in TRACK_FUNCS:
        raise ValueError(
            f'Invalid track type: {type}. Choose from: {",".join(TRACK_FUNCS.keys())}'
        )
    colorfx = green
    if command == "stop":
        colorfx = red

    # calculate total_duration of track
    total_duration = calculate_total_duration(**options)

    click.echo(
        f"▶️ {colorfx(f'{command}ing')} ➡️ {yellow('track')}: {blue(track).ljust(25)} {yellow('audio_device')}: {blue(ad or 'default').ljust(20)} {yellow('parent_pid')}: {blue(pid).ljust(15)} {yellow('total_duration')}: {blue(total_duration)}",
        err=True,
    )
    return TRACK_FUNCS.get(type)(**options)


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@seq_command_arg
@click.argument("base_config", type=click.File(), required=False)
@seq_opts
@click.pass_context
def cli(context, **kwargs):
    """
    Play a sequence of `chord`, `midi`, `note`, and/or `arp` tracks
    via yaml configuration.

    BASE_CONFIG is a path to a yaml file.

    COMMAND can be:

        play: Plays the sequence as specified in the config file

        start Starts all tracks in the sequence, irregardless of their "start" times.

        stop: Stop tracks in the sequence.

        render: Writes out individual aiff/wav files for each track in the sequence.

        echo: Pretty print the sequence config to the console.

    """
    run(**set_config_overrides_opt(context, **kwargs))



#!/usr/bin/env python
"""
This script aids in categorizing phonemes as drone-y, noise-y, or note-y for each voice
"""
from collections import defaultdict
from pprint import pprint

import click

from saysynth.constants import SAY_ALL_PHONEMES, SAY_TUNED_VOICES
from saysynth.lib import say

TEXT_PATTERN = (
    "[[inpt TUNE]] ~"
    " %s {D 800; P 65.41:0 65.41:100}"
    " %s {D 800; P 65.41:0 65.41:100}"
    " %s {D 800; P 65.41:0 65.41:100}"
    " ~ [[inpt TEXT]]"
)

CLASSIFICATIONS = {"d": "drone", "s": "noise", "e": "note", "r": "discard"}


@click.command()
def main():
    classifications = defaultdict(lambda: defaultdict(list))
    for voice in SAY_TUNED_VOICES:
        for phoneme in SAY_ALL_PHONEMES:
            text = TEXT_PATTERN % (phoneme, phoneme, phoneme)
            while 1:
                say.run(input_text=text, voice=voice, parent_pid=None)
                question = f'Does this phoneme "{phoneme}" for {voice} sound like a drone(d), noise(s), note(e), or discard(r)? Or should I repeat(t) it?'
                answer = click.prompt(
                    question, type=click.Choice(["d", "s", "e", "r", "t"])
                )
                if answer != "t":
                    # repeat the command
                    break
            classification = CLASSIFICATIONS.get(answer)
            if classification == "discard":
                click.echo(f"Okay, I'll discard {phoneme} for {voice}")
            click.echo(
                f'Got it! Recording "{phoneme}" as a {classification} for {voice}'
            )
            classifications[voice][classification].append(phoneme)
    click.echo("Done! Here are you classifications:\n")
    pprint(dict(**classifications))


if __name__ == "__main__":
    main()



"""
Play a built-in demo.
"""

import os

import click
import pkg_resources

from saysynth.cli.commands.seq import run as run_sequence
from saysynth.cli.options import (seq_command_arg, seq_opts,
                                  set_config_overrides_opt)

DEMO_DIR = pkg_resources.resource_filename("saysynth", "cli/commands/demos/")
DEMO_NAMES = {
    f.split("/")[-1]
    .split(".yml")[0][2:]
    .strip(): f.split("/")[-1]
    .split(".yml")[0]
    .split("-")[0]
    .strip()
    for f in os.listdir(DEMO_DIR)
}


def run(**kwargs):
    name = kwargs.pop("name")
    num = DEMO_NAMES.get(name)
    kwargs["base_config"] = open(
        os.path.join(DEMO_DIR, f"{num}-{name}.yml"), "r"
    )
    run_sequence(**kwargs)


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@seq_command_arg
@click.argument(
    "name", type=click.Choice(list(DEMO_NAMES.keys())), default=None
)
@seq_opts
@click.pass_context
def cli(context, **kwargs):
    """
    Play a built-in demo.
    """
    return run(**set_config_overrides_opt(context, **kwargs))



"""
The Note class creates a list of `Segment` with configurable
duration, pitch, phonemes, and volume envelope.
<center><img src="/assets/img/coffee.png"></img></center>
"""
import copy
import random
from typing import Any, Dict, List, Optional, Tuple, Union

from midi_utils import ADSR, midi_to_note, note_to_midi

from ..constants import (SAY_ALL_PHONEMES, SAY_PHONEME_CLASSES,
                         SAY_PHONEME_VOICE_CLASSES, SAY_SEGMENT_MAX_DURATION,
                         SAY_SEGMENT_SILENCE_DURATION, SAY_TUNED_VOICES)
from ..utils import bpm_to_time, frange
from .base import SayObject
from .lyrics import Lyrics
from .segment import Segment


class Note(SayObject):
    def __init__(
        self,
        note: Union[int, str] = "A3",
        phoneme: List[str] = ["m"],
        text: Optional[str] = None,
        # start position
        start: Optional[int] = 0,
        start_bpm: Optional[Union[float, int]] = 120,
        start_count: Union[str, float, int] = 0,
        start_time_sig: str = "4/4",
        # envelope
        velocity: int = 127,
        volume_level_per_segment: int = 3,
        include_volume_level: bool = True,
        attack: Union[float, int] = 0,
        decay: Union[float, int] = 0,
        sustain: Union[float, int] = 1,
        release: Union[float, int] = 0,
        # length
        duration: Optional[Union[float, int]] = None,
        duration_bpm: Optional[Union[float, int]] = 120,
        duration_count: Union[str, float, int] = 1,
        duration_time_sig: str = "4/4",
        # segmentation
        segment_duration: Optional[int] = None,
        segment_bpm: Optional[float] = 120,
        segment_count: Optional[Union[str, float, int]] = 1.0 / 8.0,
        segment_time_sig: Optional[str] = "4/4",
        # randomization
        randomize_phoneme: Optional[str] = None,
        randomize_velocity: Optional[Tuple[int, int]] = None,
        randomize_octave: Optional[List[int]] = [],
        randomize_segments: Optional[List[str]] = [],
        randomize_start: Optional[Tuple[int, int]] = None,
        **segment_options,
    ):
        f"""
        Generate say text for a collection of phonemes with adsr, pitch modulation, and more.
        Args:
            note: The note to play, eg "A3"
            phoneme: A valid combination of Phonemes documented in [Apple's Speech Synthesis guide](https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/SpeechSynthesisProgrammingGuide/Phonemes/Phonemes.html#//apple_ref/doc/uid/TP40004365-CH9-SW1).
            text: The text to "sing". If provided, this will override phoneme.
            start: The number of milliseconds of silence to add to the beginning of the track.
            start_bpm: A BPM to use when calculating the number of milliseconds of silence to add to the beginning of the track.
            start_count: A count to use when calculating the number of milliseconds of silence to add to the beginning of the track.
            start_time_sig: A time signature to use when calculating the number of milliseconds of silence to add to the beginning of the track.
            velocity: The midi velocity value to use for this note (0-127).
            volume_level_per_segment: The number of segments after which volume settings will be rendered (eg: "3" would mean one segment would have volume settings and then the next two would not, etc.))
            include_volume_level: Whether or not to the render the volume settings for this note.
                         Over-rendering these settings can lead to audio drop-outs.
            attack: A value between 0 and 1 representing the ratio of the note's total length during which the note will increase to it's amplitude.
                    A lower number is a faster attack while a larger number is a slow attack. (see `midi_utils.ADSR`).
            decay:  A value between 0 and 1 representing the ratio of note's total length during which the note will decrease in amplitude from the max amplitude to the sustain level. (see `midi_utils.ADSR`).
            sustain: A value between 0 and 1 representing the relative volume level of the sustain phase (0 is the min volume_range, 1 is the max).
            release: A value between 0 and 1 representing the ratio of the note's total length during which the note will decrease in amplitude from the sustain level to zero.
            duration: The duration of this note in number of milliseconds.
            duration_bpm: A BPM to use when calculating the note's duration.
            duration_count: A count to use when calculating the note's duration.
            duration_time_sig: A time signature to use when calculating the note's duration.
            segment_duration: The duration of each `Segment` of this note in number of milliseconds.
            segment_bpm: A BPM to use when calculating the duration of each `Segment` in this note.
            segment_count: A count to use when calculating duration of each `Segment` of this note
            segment_time_sig: A time signature to use when calculating the duration of each `Segment` in this note.
            randomize_phoneme: Randomize the phoneme for every note. "
                If "all" is passed, all valid phonemes will be used.
                Alternatively pass a list of phonemes (eg 'm,l,n') or a voice and style, eg: Fred:drone.
                Valid voices include: {', '.join(SAY_TUNED_VOICES)}.
                Valid styles include: {', '.join(SAY_PHONEME_CLASSES)}.
            randomize_velocity: Randomize a note's velocity by supplying a min and max midi velocity (eg: -rv [40, 120])
            randomize_octave: A list of octaves to randomly vary between.
                              You can weight certain octaves by providing them multiple times
                              (eg: [0,0-1,-1,2] would prefer the root octave first, one octave down second, and two octaves up third).
            randomize_segments: Randomize the 'phoneme', 'octave', and/or 'velocity' according to each respective randomization setting.
            randomize_start: Randomize the number of milliseconds to silence to add before the say text.
                             The first number passed in is the minimum of the range, the second is the max (eg: [4000, 12000] would set a range for four to twelve seconds).
            **segment_options: Additional options to pass to each `Segment`.
        """

        self.segment_options = segment_options
        root = segment_options.pop("root", None)

        if root or note:
            self.note = note_to_midi(root or note)  # root == note
            self.name = midi_to_note(self.note)
        else:
            self.note = 0
            self.name = "silence"

        # phoneme
        self.phoneme = phoneme
        if isinstance(self.phoneme, str):
            self.phoneme = [phoneme]

        # text / lyrics
        self.lyrics = None
        if text:
            self.lyrics = Lyrics(text)

        # start position
        self.start = start
        if not self.start:
            self.start = bpm_to_time(start_bpm, start_count, start_time_sig)
        if randomize_start:
            self.start = random.choice(
                range(self.randomize_start[0], self.randomize_start[1] + 1)
            )

        # duration
        self.duration = duration
        if not self.duration:
            self.duration = bpm_to_time(
                duration_bpm, duration_count, duration_time_sig
            )

        # velocity
        self.velocity = velocity
        self.volume_level_per_segment = volume_level_per_segment
        self.include_volume_level = include_volume_level

        # segmentation
        self.segment_duration = segment_duration
        if not self.segment_duration:
            self.segment_duration = bpm_to_time(
                segment_bpm, segment_count, segment_time_sig
            )
        self.segment_duration = min(
            SAY_SEGMENT_MAX_DURATION, self.segment_duration
        )
        self.segment_count = int(self.duration / self.segment_duration) + 1

        # adsr
        self.adsr = ADSR(
            attack, decay, sustain, release, samples=self.segment_count
        )

        # randomization
        self.randomize_phoneme = randomize_phoneme
        self.randomize_velocity = randomize_velocity
        self.randomize_octave = randomize_octave
        self.randomize_segments = randomize_segments
        self.randomize_start = randomize_start

    def _get_random_phoneme(self, index: int) -> str:
        if self.randomize_phoneme == "all":
            return random.choice(SAY_ALL_PHONEMES)
        elif ":" in self.randomize_phoneme:
            voice, style = self.randomize_phoneme.split(":")
            voice = voice.title()  # allow for lowercase
            try:
                return random.choice(SAY_PHONEME_VOICE_CLASSES[voice][style])
            except KeyError:
                raise ValueError(
                    f"Invalid `voice` '{voice}' or `style` '{style}'. "
                    f"`voice` must be one of: {', '.join(SAY_TUNED_VOICES)}. "
                    f"`style` must be one of: {', '.join(SAY_PHONEME_CLASSES)}"
                )
        else:
            return random.choice(
                [c.strip() for c in self.randomize_phoneme.split(",")]
            )

    def _get_phoneme(self, index: int) -> str:
        # handle phoneme randomization
        if self.randomize_phoneme:
            return self._get_random_phoneme(index)

        if self.lyrics:
            return self.lyrics.get_phoneme_for_index(index)

        return self.phoneme[index % len(self.phoneme)]

    def _get_note(self) -> int:
        if len(self.randomize_octave):
            return (random.choice(self.randomize_octave) * 12) + note_to_midi(
                self.note
            )
        return self.note

    def _get_velocity(self) -> int:
        if self.randomize_velocity:
            return random.choice(
                range(
                    self.randomize_velocity[0], self.randomize_velocity[1] + 1
                )
            )
        return self.velocity

    def _get_segment_kwargs(self, **kwargs) -> Dict[str, Any]:
        opts = copy.copy(self.segment_options)
        opts.update(kwargs)
        return opts

    def _randomize_segment(self, note, velocity):

        # optionally randomize every segment.
        if "octave" in self.randomize_segments and self.randomize_octave:
            note = self._get_note()
        if "velocity" in self.randomize_segments and self.randomize_velocity:
            velocity = self._get_velocity()
        return note, velocity

    def _get_segment(
        self,
        index: int = 0,
        note: str = None,
        velocity: int = 0,
        duration: Optional[float] = None,
        **kwargs,
    ) -> Segment:
        """
        Generate each segment of the Note, applying randomization, ADSR settings, phoneme generation, and other Segment parameters.
        """
        note, velocity = self._randomize_segment(note, velocity)
        return Segment(
            note=note,
            velocity=velocity * self.adsr.get_value(index),
            phoneme=self._get_phoneme(index),
            duration=duration or self.segment_duration,
            include_volume_level=self.include_volume_level
            and index % self.volume_level_per_segment == 0,
            **self._get_segment_kwargs(**kwargs),
        )

    @property
    def segments(self) -> List[Segment]:
        """
        The generated list of `Segment` within the note.
        """
        _segments = []
        # get initial value of note + velocity + phoneme
        note = self._get_note()
        velocity = self._get_velocity()

        if self.start and self.start > 0:
            # create multiple silent phonemes which add up to the desired start position
            start_breaks = list(
                frange(0.0, self.start, SAY_SEGMENT_SILENCE_DURATION, 10)
            )
            for index, total_start_time in enumerate(start_breaks[1:]):
                segment = self._get_segment(index, type="silence", velocity=0)
                _segments.append(segment)

            if total_start_time < self.start:
                # add final step of silence
                _segments.append(
                    self._get_segment(index + 1, type="silence", velocity=0)
                )

        # create multiple phonemes which add up to the phoneme_duration
        segment_breaks = list(
            frange(0.0, self.duration, self.segment_duration, 10)
        )
        total_time = 0
        index = 0
        for index, total_time in enumerate(segment_breaks[1:]):
            segment = self._get_segment(
                index,
                note,
                velocity,
                type=self.segment_options.get("type", "note"),
            )
            _segments.append(segment)

        if total_time < self.duration and len(_segments) < self.segment_count:

            # add final step
            _segments.append(
                self._get_segment(
                    index + 1,
                    note,
                    velocity,
                    duration=self.duration - total_time,
                    type=self.segment_options.get("type", "note"),
                )
            )
        return _segments

    @property
    def n_segments(self):
        """
        The number of Segments in the Note.
        """
        return len(self.segments)

    def to_text(self) -> str:
        """
        Render this Note as Apple SpeechSynthesis DSL text.
        """
        return "\n".join([s.to_text() for s in self.segments])



"""
Synthesize an arpeggiated melody.
"""
import click
from midi_utils import note_to_midi

from saysynth.cli.options import (adsr_opts, arp_opts, chord_opts,
                                  duration_opts, log_configurations,
                                  output_file_opt, phoneme_opts,
                                  randomize_velocity_opt, say_opts,
                                  segment_opts, start_opts, text_opt,
                                  velocity_emphasis_opt, volume_level_opts,
                                  volume_range_opt)
from saysynth.core import Arp, controller


def run(**kwargs):
    """
    Generate an arpeggiated melody.
    """
    arp = Arp(**kwargs)
    arp.cli(**kwargs)


@click.command()
@click.argument("root", type=note_to_midi, default="A2")
@arp_opts
@chord_opts
@velocity_emphasis_opt
@volume_range_opt
@randomize_velocity_opt
@volume_level_opts
@start_opts
@duration_opts
@phoneme_opts
@text_opt
@segment_opts
@adsr_opts
@output_file_opt
@say_opts
def cli(**kwargs):
    """
    Generate an arpeggiated melody.
    """
    if kwargs["yaml"]:
        return log_configurations("arp", **kwargs)
    kwargs = controller.handle_cli_options("arp", **kwargs)
    return run(**kwargs)



"""
A Python-wrapper for Apple's [`say`](https://ss64.com/osx/say.html) command.
"""

import os
import subprocess
import warnings
from typing import Any, Dict, List, Optional

from ..constants import (SAY_BIG_ENDIAN_ONLY_FILE_FORMATS, SAY_COLORS,
                         SAY_DATA_TYPES, SAY_DEFAULT_FLOAT_SAMPLE_SIZE,
                         SAY_ENDIANNESS, SAY_EXECUTABLE, SAY_FILE_FORMATS,
                         SAY_MAX_SAMPLE_RATE, SAY_SAMPLE_SIZES,
                         SAY_VALID_FLOAT_SAMPLE_SIZES)
from ..core import controller
from ..utils import make_tempfile


def _gen_data_format_arg(
    file_format: str,
    endianness: str,
    data_type: str,
    sample_size: int,
    sample_rate: int,
):
    """
    Generate a string to pass to --data-format
    """
    if endianness not in SAY_ENDIANNESS:
        raise ValueError(
            "Invalid `endianess`. Choose from: LE (little endian) or BE (big endian)"
        )
    if data_type not in SAY_DATA_TYPES:
        raise ValueError(
            "Invalid `data_type`. Choose from: F (float), I (integer), UI (unsigned integer)"
        )
    if sample_size not in SAY_SAMPLE_SIZES:
        raise ValueError(
            f'Invalid `sample_size`. Choose from: {", ".join(SAY_SAMPLE_SIZES)}'
        )

    # allow pass passing sample rate as small number (eg: 24 -> 24000)
    if sample_rate < 1000:
        sample_rate *= 1000

    # don't allow a sample rate greater than the maximum
    if sample_rate > SAY_MAX_SAMPLE_RATE:
        sample_rate = SAY_MAX_SAMPLE_RATE

    # big endian-only formats:
    if file_format in SAY_BIG_ENDIAN_ONLY_FILE_FORMATS and file_format != "BE":
        msg = (
            f"file_format '{file_format}' only accepts and endianness of 'BE'"
        )
        warnings.warn(msg, SyntaxWarning)
        endianness = "BE"

    # check sample size by data_type
    if data_type == "F" and sample_size not in SAY_VALID_FLOAT_SAMPLE_SIZES:
        msg = f"data_type 'F' only accepts sample_sizes of '32' and '64', setting '{sample_size}' to '{SAY_DEFAULT_FLOAT_SAMPLE_SIZE}'"
        warnings.warn(msg, SyntaxWarning)
        sample_size = SAY_DEFAULT_FLOAT_SAMPLE_SIZE

    return f"{endianness}{data_type}{sample_size}@{int(sample_rate)}"


def _gen_interactive_arg(text_color: str = "white", bg_color: str = "black"):
    """
    Generate a string to pass to --interactive
    """
    if bg_color and not text_color:
        text_color = (
            "white"  # default text color if only background is supplied
        )
    if text_color not in SAY_COLORS:
        raise ValueError(
            f'Invalid `text_color`, choose from: {", ".join(SAY_COLORS)}'
        )
    if bg_color not in SAY_COLORS:
        raise ValueError(
            f'Invalid `bg_color`, choose from: {", ".join(SAY_COLORS)}'
        )
    return f"--interactive={text_color}/{bg_color}"


def cmd(
    input_text: Optional[str] = None,
    voice: Optional[str] = None,
    rate: Optional[int] = None,
    input_file: Optional[str] = None,
    audio_output_file: Optional[str] = None,
    service_name: Optional[str] = None,
    network_send: Optional[str] = None,
    audio_device: Optional[str] = None,
    stereo: bool = False,  # whether or not
    endianness: str = "LE",  # LE/BE
    data_type: str = "I",  # F/I/UI
    sample_size: Optional[int] = 8,
    sample_rate: Optional[int] = 22050,
    quality: int = 127,
    progress: bool = False,
    interactive: bool = False,
    text_color: Optional[str] = None,
    bg_color: Optional[str] = None,
    executable: str = SAY_EXECUTABLE,
    **kwargs,
) -> List[str]:
    """
    A python wrapper around the say command.

    Args:
        input_text: The text to speak
        voice: Specify the voice to be used. Default is the voice selected in System Preferences. To obtain a list of voices installed in the system, specify "?" as the voice name.
        rate:  Speech rate to be used, in words per minute.
        input_file:  Specify a file to be spoken. If file is - or neither this parameter nor a message is specified, read from standard input.
        audio_output_file: Specify the path for an audio file to be written. AIFF is the default and should be supported for most voices, but some
        voices support many more file formats.
        service_name:  Specify a service name (default "AUNetSend")
        network_send: Specify an IP and port to be used for redirecting the speech output through AUNetSend.
        audio_device: Specify, by ID or name prefix, an audio device to be used to play the audio. To obtain a list of audio output devices, specify "?" as the device name.
        stereo: Whether or not to output a stereo signal
        endianness: str = "LE",  # LE/BE
        data_type: str = "F",  # F/I/U
        sample_size: One of 8, 16, 24, 32, 64.
        sample_rate: Optional[int] = 22050,
        quality: The audio converter quality level between 0 (lowest) and 127 (highest).
        progress: Display a progress meter during synthesis.
        interactive: Print the text line by line during synthesis, highlighting words as they are spoken. Markup can be one of:
            * A terminfo capability as described in terminfo(5), e.g. bold, smul, setaf 1.:
            * A color name, one of black, red, green, yellow, blue, magenta, cyan, or white.:
            * A foreground and background color from the above list, separated by a slash, e.g. green/black. If the foreground color is omitted, only the background color is set.:
            * If markup is not specified, it defaults to smso, i.e. reverse video.:
            * If the input is a TTY, text is spoken line by line, and the output file, if specified, will only contain audio for the last line of the input.  Otherwise, text is spoken all at once.
        text_color: A color name, one of black, red, green, yellow, blue, magenta, cyan, or white.
        bg_color: A color name, one of black, red, green, yellow, blue, magenta, cyan, or white.
        executable: The path to the `say` executable (default '/usr/bin/say')

    """  # noqa: E501
    if not input_text and not input_file:
        raise ValueError("Must provide `input_text` or `input_file`")

    # override text if input file is provided
    if input_file:
        # verify that input file exists
        if not os.path.exists(input_file):
            raise ValueError("`input_file`: {input_file} does not exist!")

    # verify quality
    if quality < 0 or quality > 127:
        raise ValueError("`quality` must be between 0 and 127")

    # construct base command
    cmd = [executable]
    if input_text:
        cmd.append(input_text)
    elif input_file:
        cmd.extend(["-f", input_file])
    if voice:
        cmd.extend(["-v", voice])
    if rate:
        cmd.extend(["-r", rate])

    if audio_output_file:
        # verify file_format:
        extension = audio_output_file.lower().split(".")[-1]
        if extension not in SAY_FILE_FORMATS:
            raise ValueError(
                f"Invalid extension: '.{extension}'. Choose from: {', '.join(SAY_FILE_FORMATS.keys())}"
            )
        file_format = SAY_FILE_FORMATS.get(extension)
        cmd.extend(["--file-format", file_format])
        cmd.extend(["-o", audio_output_file])
        data_format = _gen_data_format_arg(
            file_format, endianness, data_type, sample_size, sample_rate
        )
        cmd.extend(["--data-format", data_format])

        if stereo:
            cmd.append("--channels=2")
    else:
        cmd.append(f"--quality={quality}")
        # handle network output if output file is not specified
        if service_name:
            cmd.extend(["-n", service_name])
        if network_send:
            cmd.extend(f"--network-send={network_send}")
        if audio_device:
            cmd.extend(["-a", audio_device])

    # progress bar
    if progress:
        cmd.append("--progress")

    # interactivity
    if interactive:
        cmd.append(_gen_interactive_arg(text_color, bg_color))
    args = [str(a) for a in cmd]
    # TODO: setup debug logging
    # msg = f"Executing say command:\n$ {' '.join(args)}"
    # print(msg)
    return args


def run(args: Optional[List] = None, **kwargs) -> None:
    """
    Execute a command given a list of arguments outputted by `cmd`
    or by supplying the kwargs that `cmd` accepts
    Args:
        args: A list of args generated by `cmd`
    """
    wait_for_process = kwargs.pop("wait", False)
    if not args:
        parent_pid = kwargs.pop("parent_pid", os.getpid())
        parent_pid_file = kwargs.pop("parent_pid_file", None)
        # write text as input file
        tempfile = None
        if kwargs.get("input_text") and not kwargs.get("input_file", None):
            text = kwargs.pop("input_text")
            # override text with tempfile
            tempfile = make_tempfile()
            with open(tempfile, "w") as f:
                f.write(text)
            kwargs["input_file"] = tempfile
        args = cmd(**kwargs)
    process = None
    try:
        process = subprocess.Popen(args, stdout=subprocess.PIPE)
        # register each process with the parent pid.
        controller.add_child_pid(process.pid, parent_pid, parent_pid_file)
        if wait_for_process:
            process.wait()
    except KeyboardInterrupt:
        pass


def _run_spawn(kwargs: Dict[str, Any]) -> None:
    """
    Utility for passing kwargs into `run` within `spawn`
    """
    return run(**kwargs)


def spawn(commands: List[Dict[str, Any]]) -> None:
    """
    Spawn multiple say processes in parallel by
    passing in a list of commands generated by `cmd`

    Args:
        commands: A list of command args generated by `cmd`
    """
    for command in commands:
        _run_spawn(command)



"""
Stop currently running `saysynth` processes by `sequences`, `tracks`, `audio_devices`, and/or `parent_pids`
"""
import click

from saysynth.cli.colors import blue, red, yellow
from saysynth.core import controller


@click.command()
@click.option(
    "-p",
    "--pids",
    "parent_pid",
    type=lambda x: [int(t.strip()) for t in x.split(",") if t.strip()],
    help="Stop currently running `saysynth` processes by passing in the `parent_pids`",
    default=None,
)
@click.option(
    "-s",
    "--sequences",
    "seq",
    type=lambda x: [t.strip() for t in x.split(",") if t.strip()],
    help="Stop currently running `saysynth` processes by passing in the `sequence ` names",
    default=None,
)
@click.option(
    "-ad",
    "--audio-devices",
    "ad",
    type=lambda x: [t.strip() for t in x.split(",") if t.strip()],
    help="Stop currently running `saysynth` processes by passing in the `audio_devices`",
    default=None,
)
@click.option(
    "-t",
    "--tracks",
    "track",
    type=lambda x: [t.strip() for t in x.split(",") if t.strip()],
    help="Stop currently running `saysynth` processes by passing in the `track` names",
    default=None,
)
def cli(**kwargs):
    """
    Stop currently running `say` processes by `sequence`, `track`, `audio_device`, and/or `parent_pid`
    """
    all_null = True
    for key in ["seq", "track", "ad", "parent_pid"]:
        vals = kwargs.get(key, None)
        if not vals:
            continue
        all_null = False
        for val in vals:
            click.echo(f"🛑 {red('stopping')} ➡️ {yellow(key)}: {blue(val)}")
            controller.stop_child_pids(**{key: val})
            if key == "parent_pid":
                controller.rm_parent_pid(parent_pid=val)
    if all_null:
        click.echo(f"🛑 {red('stopping')} ➡️ {yellow('all processes')}!")
        controller.stop_child_pids()



"""
A `Segment` is the base unit text according to
[Apple's Speech Synthesis Programming Guide](https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/SpeechSynthesisProgrammingGuide/SpeechOverview/SpeechOverview.html#//apple_ref/doc/uid/TP40004365-CH3-SW1).
A segment represents an individual part-of-speech (phoneme) in `say` and can have
one or more phonemes, a duration, and a sequences of pitches.

Segments are combined together throughout `saysynth` to create musical passages.

<center><img src="/assets/img/cell.png"></img></center>

"""
from typing import Tuple, Union

from midi_utils import midi_to_freq, note_to_midi

from ..constants import (SAY_EMPHASIS, SAY_PHONEME_SILENCE,
                         SAY_SEGMENT_MAX_DURATION, SAY_VOLUME_RANGE)
from ..utils import rescale


class Segment(object):
    def __init__(
        self,
        note: Union[int, str],
        velocity: int = 127,
        phoneme: str = "m",
        duration: Union[float, int] = SAY_SEGMENT_MAX_DURATION,
        type: str = "note",
        emphasis: Tuple[int, int] = SAY_EMPHASIS,
        volume_range: Tuple[float, float] = SAY_VOLUME_RANGE,
        include_volume_level: bool = True,
        duration_sig_digits: int = 4,
        **kwargs,
    ):
        """
        An individual segment of speech in Apple's DSL
        Args:
            note: The note to map to a frequency, eg "A3"
            velocity: The midi velocity value to use for the segment (0-127).
            phoneme:  A valid combination of Phonemes documented in [Apple's Speech Synthesis guide](https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/SpeechSynthesisProgrammingGuide/Phonemes/Phonemes.html#//apple_ref/doc/uid/TP40004365-CH9-SW1).
            duration: The duration of the segment in milliseconds.
            type: Either "note" or "silence"
            emphasis: A level of emphasis to place on this segment (0,1, or 2)
            volume_range: A range between 0 and 127 representing the minimum and maximum velocity values to render.
            include_volume_level: Whether or not to the render the volume settings for this segment.
                         Over-rendering these settings can lead to audio drop-outs.
            duration_sig_digits: The number of significant digits to use when rendering the duration value.
                                 A higher value should yield more rhythmically precise results.
        """
        self._phoneme = phoneme
        self._duration = duration
        self._emphasis = emphasis
        self.velocity = velocity
        self.note = note
        self.is_silence = type == "silence"
        self.volume_range = volume_range
        self.include_volume_level = include_volume_level
        self.duration_sig_digits = duration_sig_digits

    @property
    def phoneme(self):
        return self._phoneme

    @property
    def phoneme_has_emphasis(self) -> bool:
        """
        Return True if the `phoneme` has an included
        emphasis.
        """
        if (
            self._phoneme.startswith("0")
            or self._phoneme.startswith("1")
            or self._phoneme.startswith("2")
        ):
            return True
        return False

    @property
    def frequency_envelope(self) -> str:
        """
        Translate a note name to a frequency.
        **TODO:** Add intra-note modulation.
        """
        freq = midi_to_freq(note_to_midi(self.note))
        return f"P {freq}:0"

    @property
    def duration(self) -> float:
        """
        Clamp segment duration at `SAY_SEGMENT_MAX_DURATION`
        and round it to `self.duration_sig_digits`
        """
        return round(
            min(self._duration, SAY_SEGMENT_MAX_DURATION),
            self.duration_sig_digits,
        )

    @property
    def volume(self) -> str:
        """
        Translate a midi velocity value (0-127) into a pair
        of say volume tags, eg: "[[ volm +0.1 ]]"
        """
        if self.include_volume_level:
            volume = rescale(self.velocity, [0, 127], self.volume_range)
            return f"[[ volm {volume} ]]"
        return ""

    @property
    def emphasis(self) -> str:
        """
        Translate a midi velocity value (0-127) into a phoneme
        emphasis value ("", "1", or "2")when provided with a tuple
        of steps (step_1, step_2) eg: (75, 100)
        This action is not performed when the phoneme already
        has an emphasis included.
        """
        if not self.phoneme_has_emphasis:
            if not self.velocity:
                return ""
            if self.velocity > self._emphasis[1]:
                return "2"
            if self.velocity > self._emphasis[0]:
                return "1"
        return ""

    def to_text(self) -> str:
        """
        Render this Segment as Apple SpeechSynthesis DSL text.
        """
        if self.is_silence:
            return f"{self.volume} {SAY_PHONEME_SILENCE} {{D {self.duration}}}"
        return f"{self.volume} {self.emphasis}{self.phoneme} {{D {self.duration}; {self.frequency_envelope}}}"

    def __eq__(self, other):
        return self.to_text() == other.to_text()

    def __str__(self) -> str:
        return self.to_text()



"""
List all currently running `saysynth` processes.
"""
import click

from saysynth.cli.colors import blue, green, red, yellow
from saysynth.core import controller


def run(**kwargs):
    pids = controller.list_pids()
    if not len(pids):
        click.echo(red("There are no active processes!"))
        return
    last_seq = None
    for p in pids:
        seq = p["seq"]
        if last_seq != seq:
            click.echo(r"-" * 79, err=True)
            click.echo(f"{red('sequence')}: {green(seq)}")
            click.echo(r"-" * 79, err=True)
        click.echo(
            f"➡️ {yellow('track')}: {blue(p['track']).ljust(23)} {yellow('audio_device')}: {blue(p['ad']).ljust(18)} {yellow('parent_pid')}: {blue(p['parent_pid']).ljust(14)} {yellow('child_pids')}: {blue(len(p['child_pids']))}"
        )
        last_seq = seq


@click.command()
def cli(**kwargs):
    """
    List all currently running `saysynth` processes.
    """
    run(**kwargs)



from pathlib import Path

from setuptools import find_packages, setup

from saysynth.version import VERSION

config = {
    "name": "saysynth",
    "version": VERSION,
    "packages": find_packages(),
    "include_package_data": True,
    "package_data": {
        "": ["cli/commands/demos/.*.yml", "*.md"],
    },
    "install_requires": [
        "charset-normalizer",
        "click",
        "mido",
        "midi-utils",
        "pyyaml",
        "g2p_en",
        "nltk",
        "hashids",
    ],
    "author": "Brian Abelson",
    "author_email": "hey@gltd.email",
    "description": "Make music with Mac's say command",
    "long_description": (Path(__file__).parent / "README.md").read_text(),
    "long_description_content_type": 'text/markdown',
    "url": "http://saysynth.org",
    "entry_points": {
        "console_scripts": [
            "saysynth=saysynth.cli:main",
            "sy=saysynth.cli:main",
        ]
    },
}

setup(**config)



"""
Constants for use throughout `saysynth`
"""
import os

SAY_EXECUTABLE = os.getenv("SAYSYNTH_SAY_EXECUTABLE", "/usr/bin/say")
"""The path to where the `say` command exists."""

SAY_FILE_FORMATS = ["wav", "aiff"]
"""Valid output file formats."""

SAY_TUNED_VOICES = ["Alex", "Fred", "Victoria"]
"""`say` voices which respect [[TUNE]] input"""

SAY_TUNE_TAG = "[[inpt TUNE]]"
"""Opening tag for tuned input to `say`"""

SAY_COLORS = [
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
]
"""Colors which can used to style the interactive output of `say`"""

SAY_ENDIANNESS = ["BE", "LE"]
"""Values for --data-format to determine the endianness"""

SAY_DATA_TYPES = ["F", "I", "UI"]
"""Values for --data-format to determine the data type"""

SAY_SAMPLE_SIZES = [8, 16, 24, 32, 64]
"""Values for --data-format to determine the sample size"""

SAY_FILE_FORMATS = {
    "mp3": "mp4f",
    "aiff": "AIFF",
    "aif": "AIFF",
    "flac": "FLAC",
    "m4a": "m4af",
    "wav": "WAVE",
}
"""Mapping of file extension to say's --file-format arguments"""

SAY_BIG_ENDIAN_ONLY_FILE_FORMATS = ["AIFF", "m4af", "FLAC"]
"""List of say's --file-format arguments which accept only BE endianness"""

SAY_DEFAULT_FLOAT_SAMPLE_SIZE = 32
"""The default sample size to use when the data_type is F"""

SAY_VALID_FLOAT_SAMPLE_SIZES = [32, 64]
"""A list of valid sample sizes when the data_type is F"""

SAY_MAX_SAMPLE_RATE = 22050
"""
This is the max sample rate,
anything above this [will generate up-sampled audio](https://stackoverflow.com/questions/9729153/error-on-say-when-output-format-is-wave)
"""

SAY_ALL_PHONEMES = [
    "AE",
    "EY",
    "AO",
    "AX",
    "IY",
    "EH",
    "IH",
    "AY",
    "IX",
    "AA",
    "UW",
    "UH",
    "UX",
    "OW",
    "AW",
    "OY",
    "b",
    "C",
    "d",
    "D",
    "f",
    "g",
    "h",
    "J",
    "k",
    "l",
    "m",
    "n",
    "N",
    "p",
    "r",
    "s",
    "S",
    "t",
    "T",
    "v",
    "w",
    "y",
    "z",
    "Z",
]
"""
[A list of all valid phonemes to pass into `say`]( https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/SpeechSynthesisProgrammingGuide/Phonemes/Phonemes.html#//apple_ref/doc/uid/TP40004365-CH9-SW1).
"""

SAY_PHONEME_CLASSES = ["drone", "noise", "note"]
""" Classes of phonemes as defined in scripts/classify_phonemes.py """

SAY_PHONEME_VOICE_CLASSES = {
    "Alex": {
        "drone": [
            "AE",
            "EY",
            "AO",
            "AX",
            "IY",
            "EH",
            "IH",
            "AY",
            "IX",
            "UW",
            "OW",
            "OY",
            "h",
            "l",
            "m",
            "n",
            "N",
            "r",
            "Z",
        ],
        "noise": [
            "C",
            "d",
            "D",
            "f",
            "J",
            "k",
            "p",
            "s",
            "S",
            "t",
            "T",
            "v",
            "z",
            "Z",
        ],
        "note": ["AA", "UH", "UX", "AW", "b", "g", "w", "y", "Z"],
    },
    "Fred": {
        "drone": [
            "AE",
            "EY",
            "AO",
            "AX",
            "IY",
            "EH",
            "IH",
            "AY",
            "IX",
            "AA",
            "UW",
            "UH",
            "UX",
            "OW",
            "AW",
            "OY",
            "D",
            "l",
            "m",
            "n",
            "N",
            "r",
            "v",
            "w",
            "y",
            # "z",
            # "Z",
        ],
        "note": [  # all of Fred's drones work as notes.
            "AE",
            "EY",
            "AO",
            "AX",
            "IY",
            "EH",
            "IH",
            "AY",
            "IX",
            "AA",
            "UW",
            "UH",
            "UX",
            "OW",
            "AW",
            "OY",
            "D",
            "l",
            "m",
            "n",
            "N",
            "r",
            "v",
            "w",
            "y",
            "z",
            "Z",
        ],
        "noise": [
            "b",
            "C",
            "d",
            "f",
            "g",
            "h",
            "J",
            "k",
            "p",
            "s",
            "S",
            "t",
            "T",
        ],
    },
    "Victoria": {
        "drone": [
            "AE",
            "EY",
            "AO",
            "AX",
            "IY",
            "EH",
            "IH",
            "AY",
            "IX",
            "AA",
            "UW",
            "UH",
            "UX",
            "AW",
            "OY",
            "l",
            "m",
            "n",
            "N",
            "r",
            "v",
            "w",
            "y",
        ],
        "noise": [
            "C",
            "d",
            "D",
            "f",
            "h",
            "k",
            "p",
            "s",
            "S",
            "t",
            "T",
            "z",
            "Z",
        ],
        "note": ["OW", "b", "g", "J"],
    },
}
""" Generated by running scripts/classify_phonemes.py """

SAY_PHONEME_SILENCE = "%"
""" This is the sound of silence """

G2P_PHONEMES_TO_SAY_PHONEMES = {
    "AA0": "AA",
    "AA1": "1AA",
    "AA2": "2AA",
    "AE0": "AE",
    "AE1": "1AE",
    "AE2": "2AE",
    "AH0": "AAh",
    "AH1": "1AAh",
    "AH2": "2AAh",
    "AO0": "AO",
    "AO1": "1AO",
    "AO2": "2AO",
    "AW0": "AW",
    "AW1": "1AW",
    "AW2": "2AW",
    "AY0": "AY",
    "AY1": "1AY",
    "AY2": "2AY",
    "B": "b",
    "CH": "C",
    "D": "d",
    "DH": "T",
    "EH0": "EH",
    "EH1": "1EH",
    "EH2": "2EH",
    "ER0": "AXr",
    "ER1": "1AXr",
    "ER2": "2AXr",
    "EY0": "EY",
    "EY1": "1EY",
    "EY2": "2EY",
    "F": "f",
    "G": "g",
    "HH": "h",
    "IH0": "IH",
    "IH1": "1IH",
    "IH2": "2IH",
    "IY0": "IY",
    "IY1": "1IY",
    "IY2": "2IY",
    "JH": "J",
    "K": "k",
    "L": "l",
    "M": "m",
    "N": "n",
    "NG": "N",
    "OW0": "OW",
    "OW1": "1OW",
    "OW2": "2OW",
    "OY0": "OY",
    "OY1": "1OY",
    "OY2": "2OY",
    "P": "p",
    "R": "r",
    "S": "s",
    "SH": "S",
    "T": "t",
    "TH": "T",
    "UH0": "UH",
    "UH1": "1UH",
    "UH2": "2UH",
    "UW": "UW",
    "UW0": "0UW",
    "UW1": "1UW",
    "UW2": "2UW",
    "V": "v",
    "W": "w",
    "Y": "y",
    "Z": "z",
    "ZH": "Z",
}
"""
A lookup between phonemes in [G2P](https://github.com/Kyubyong/g2p/blob/master/g2p_en/g2p.py#L55)
and [say](https://developer.apple.com/library/archive/documentation/UserExperience/Conceptual/SpeechSynthesisProgrammingGuide/Phonemes/Phonemes.html#//apple_ref/doc/uid/TP40004365-CH9-SW1).
"""

SAY_SEGMENT_MAX_DURATION = 1200
"""
The number of milliseconds at which the duration
of an individual phoneme stops changing
"""

SAY_SEGMENT_SILENCE_DURATION = 1000
"""
This is the number of milliseconds to use for an individual segment of silence.
"""

SAY_EMPHASIS = [75, 100]
"""
The midi velocity values above which we add an emphasis to a phoneme.
"""

SAY_VOLUME_RANGE = [0.0, 1.0]
"""
The min and max range of volume levels to map to from midi velocities.
"""

SAY_VOLUME_LEVEL_PER_NOTE = 2
"""
The number of notes per sequence to show volume tags.
Including too many volume tags in a single command can cause random drop-outs.
"""

SAY_VOLUME_LEVEL_PER_SEGMENT = 4
"""
The number of segments per note to show volume tags.
Including too many volume tags in a single command can cause random drop-outs.
"""

DEFAULT_SEQUENCE_NAME = "sy"
"""
The sequence name to assign to a process when launched outside the context
of a sequence.
"""

DEFAULT_BPM_TIME_SIG = "4/4"
DEFAULT_BPM_TIME_BPM = 120
DEFAULT_BPM_TIME_COUNT = 1

SAY_EXTRA_OPTION_DELIMITER = "__"



"""
The `SayObject` class provides standard runtime functionality to monotonic features like `arp`, `note`, and `midi_track`.
<center><img src="/assets/img/nuclear.png"></img></center>
"""
import sys

from ..cli.options import prepare_options_for_say
from ..constants import SAY_TUNE_TAG
from ..lib import say


class SayObject(object):
    def to_text(self):
        """Render this object as Apple SpeechSynthesis DSL text"""
        raise NotImplementedError

    def to_say_text(self):
        """
        Render this object as Apple SpeechSynthesis DSL text,
        including the initial [[TUNE]] tag.
        """
        return SAY_TUNE_TAG + "\n" + self.to_text()

    def write(self, output_file) -> None:
        """
        Render this Arp as Apple SpeechSynthesis DSL text,
        and write it to an output file
        """
        with open(output_file, "w") as f:
            f.write(self.to_say_text())

    def play(self, **say_options) -> None:
        """
        Play this object and pass in additional options to `say.run`
        """
        text = self.to_say_text()
        options = prepare_options_for_say(text, **say_options)
        say.run(**options)

    @classmethod
    def new(cls, **kwargs):
        """
        Instantiate this object.
        """
        return cls(**kwargs)

    def cli(self, **kwargs) -> None:
        """
        Handle the execution of this object
        within the context of the CLI.
        """
        klass = self.new(**kwargs)
        output_file = kwargs.get("output_file")
        if output_file:
            klass.write(output_file)
        elif kwargs.get("no_exec", False):
            sys.stdout.write(klass.to_say_text())
        else:
            klass.play(**kwargs)



"""
The Lyric class enables "singing"-like speech synthesis
by translating input text into a list of phonemes which can
mapped onto musical passages through other `saysynth` functions.
<center><img src="/assets/img/cell.png"></img></center>
"""
from functools import cached_property
from typing import List

from ..lib import nlp
from .word import Word


class Lyrics(object):
    def __init__(self, text: str):
        """
        The Lyric class enables "singing"-like speech synthesis
        by translating input text into a list of phonemes which can
        mapped onto musical passages through other `saysynth` functions.
        Args:
            text: The text of the lyrics. Use punctuation for pauses.
        """
        self.text = text

    def get_phoneme_for_index(self, index: int) -> str:
        """
        Given an index (for instance in an iteration) return
        a phoneme at the respective position. This operation
        uses a modulo such that a text is repeated over and
        over until the full musical passage has been generated.
        """
        seg_index = index % self.n_phonemes
        return self.phonemes[seg_index]

    def get_phonemes(self, start_at: int = 0, end_at: int = -1) -> List[str]:
        """
        Get phonemes of these lyrics between a specific index range.
        """
        return self.phonemes[start_at:end_at]

    @cached_property
    def words(self) -> List[Word]:
        """
        A list of `Word` objects in the text.
        """
        return [Word(*word) for word in nlp.process_text_for_say(self.text)]

    @cached_property
    def n_words(self) -> int:
        """
        The number of syllables in the text.
        """
        return len(self.words)

    @cached_property
    def n_syllables(self) -> int:
        """
        The number of syllables in the text.
        """
        return sum([word.syllables for word in self.words])

    @cached_property
    def phonemes(self) -> List[str]:
        """
        The list of phonemes in the text.
        """
        return [p for word in self.words for p in word.phonemes]

    @cached_property
    def n_phonemes(self) -> int:
        """
        The number of phonemes in the text.
        """
        return len(self.phonemes)



"""
The `saysynth.cli.commands` module includes submodules containing the logic for each
saysynth command.

Each submodule defines a `run` function which contains the logic
for executing the command and a `cli` function which includes the `click` decorators.

The `seq` submodule imports the `run` functions from a number of other submodules
as it provides an interface for running multiple commands at once.

<center><img src="/assets/img/sun-wavy.png"></img></center>
"""



"""
Synthesize a polyphonic chord.
"""
import click
from midi_utils import note_to_midi

from saysynth.cli.options import (adsr_opts, chord_opts, duration_opts,
                                  log_configurations, phoneme_opts, say_opts,
                                  segment_opts, start_opts, text_opt,
                                  velocity_opts, volume_level_opts)
from saysynth.core import Chord, controller


def run(**kwargs):
    """
    Given a note name (or midi note number), stream text required to generate a continuous drone for input to say
    """
    # generate chord
    chord = Chord(**kwargs)
    chord.cli(**kwargs)


@click.command()
@click.argument("root", type=note_to_midi, default="A2")
@text_opt
@chord_opts
@start_opts
@duration_opts
@velocity_opts
@volume_level_opts
@adsr_opts
@phoneme_opts
@segment_opts
@click.option(
    "-o",
    "--output-file",
    type=str,
    help="A filepath to write the generated text to. This filepath will be used as a pattern to generate multiple text files, one per note in the chord.",
)
@say_opts
def cli(**kwargs):
    """
    Generate a polyphonic chord.
    """
    if kwargs["yaml"]:
        return log_configurations("chord", **kwargs)
    kwargs = controller.handle_cli_options("chord", **kwargs)
    return run(**kwargs)



from saysynth.lib import say


def test_gen_data_format_arg():
    data_format = say._gen_data_format_arg("AIFF", "LE", "F", 32, 22050)
    assert data_format == "BEF32@22050"
    data_format = say._gen_data_format_arg("WAVE", "LE", "F", 32, 22050)
    assert data_format == "LEF32@22050"
    data_format = say._gen_data_format_arg("WAVE", "LE", "F", 16, 22050)
    assert data_format == "LEF32@22050"


def test_say_cmd():
    args = say.cmd(
        voice="Fred",
        rate=70,
        input_text="Hello World",
        audio_device="test",
        executable="/usr/bin/say",
        stereo=True,
    )
    assert args == [
        "/usr/bin/say",
        "Hello World",
        "-v",
        "Fred",
        "-r",
        "70",
        "--quality=127",
        "-a",
        "test",
    ]
    args = say.cmd(
        voice="Alex",
        rate=70,
        input_text="Hello World",
        audio_device="test",
        executable="/usr/bin/say",
        stereo=True,
        sample_rate=12000,
        sample_size=8,
        endianness="BE",
        data_type="I",
        audio_output_file="test.wav",
    )
    assert args == [
        "/usr/bin/say",
        "Hello World",
        "-v",
        "Alex",
        "-r",
        "70",
        "--file-format",
        "WAVE",
        "-o",
        "test.wav",
        "--data-format",
        "BEI8@12000",
        "--channels=2",
    ]



"""
The Chord class creates a list of `Note` which can be played in parallel, producing polyphonic sounds.
<center><img src="/assets/img/coffee.png"></img></center>
"""
from typing import Any, Dict, List

from midi_utils import midi_chord, note_to_midi

from ..cli.options import prepare_options_for_say
from ..lib import say
from .note import Note


class Chord(object):
    """
    The Chord class creates a list of `Note` objects which can
    be played in parallel, producing polyphonic sounds.

    Args:
        note/root: The base root of the chord
        chord_notes: An optional list of midi numbers to build a chord from based off of the root. For example, the notes '[0,3,7]' with the root of 'C1' would create a C-minor chord.
        **note_options: Additional options to pass to `Note`. These will affect how each note of the arpeggiator sounds.
    """

    def __init__(self, note=None, root=None, chord_notes=[], **note_options):
        self.root = note or root or "A1"  # root == note
        self.root_midi = note_to_midi(self.root)
        self.chord_notes = chord_notes
        self.note_options = note_options

    def _get_kwargs(self, **kwargs) -> Dict[str, Any]:
        """
        get kwargs + update with new ones
        used for mapping similar kwargs over different notes
        """
        d = dict(self.note_options.items())
        d.update(kwargs)
        return d

    @property
    def midi_notes(self) -> List[int]:
        # check for arbitrary notes
        if len(self.chord_notes):
            return [self.root_midi + n for n in self.chord_notes]
        return midi_chord(root=self.root, **self.note_options)

    @property
    def notes(self) -> List[Note]:
        """
        The generated list of `Note` in the Chord.
        """
        _notes = []
        for n in self.midi_notes:
            _notes.append(Note(**self._get_kwargs(note=n)))
        return _notes

    @property
    def n_notes(self) -> int:
        """
        The number of Notes in the Chord.
        """
        return len(self.notes)

    def write(self, output_file: str) -> None:
        """
        Write each Note in the Chord to an output file,
        adding the note name to the provided filepath.
        """
        for note in self.notes:
            fn = ".".join(output_file.split(".")[:-1])
            ext = (
                "txt" if "." not in output_file else output_file.split(".")[-1]
            )
            note_output_file = f"{fn}-{note.note}.{ext}"
            with open(note_output_file, "w") as f:
                f.write(note.to_say_text())

    def _get_audio_output_file_for_note(self, filename, note: Note) -> str:
        fn = ".".join(filename.split(".")[:-1])
        ext = "aiff" if "." not in filename else filename.split(".")[-1]
        return f"{fn}-{note.name}.{ext}"

    def _get_cmd_for_note(self, note: Note):
        cmd = prepare_options_for_say(
            input_text=note.to_say_text(),
            **self._get_kwargs(note=note.note, type="note"),
        )
        audio_output_file = cmd.get("audio_output_file", None)
        if audio_output_file:
            cmd["audio_output_file"] = self._get_audio_output_file_for_note(
                audio_output_file, note
            )
        return cmd

    @property
    def commands(self) -> List[Dict[str, Any]]:
        """
        A list of note commands to spawn.
        """
        return [self._get_cmd_for_note(note) for note in self.notes]

    def play(self, **kwargs) -> None:
        """
        Play each `Note` of the Chord in parallel.
        """
        say.spawn(self.commands)

    def cli(self, **kwargs) -> None:
        """
        Handle the execution of this Chord
        within the context of the CLI.
        """
        chord = Chord(**kwargs)
        output_file = kwargs.get("output_file")
        if output_file:
            chord.write(output_file)
        else:
            chord.play(**kwargs)
